/* Parallel Hierarchical Grid -- an adaptive finite element library.
 *
 * Copyright (C) 2005-2008 State Key Laboratory of Scientific and
 * Engineering Computing, Chinese Academy of Sciences. */

/* This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA */

/* $Id: eigen.c,v 1.67 2008/01/19 07:03:32 zlb Exp $ */

#include "phg.h"
#include <stdlib.h>
#include <string.h>
#include <math.h>

/* The first eigen value */
static FLOAT lambda0 = 3. * M_PI * M_PI;

/* analytic solution of the first eigen vector in the case of the unit cube */
static void
func_u(FLOAT x, FLOAT y, FLOAT z, FLOAT *value)
{
    *value = Sin(M_PI * x) * Sin(M_PI * y) * Sin(M_PI * z);
} 

FLOAT
compute_error(DOF *u_h)
{
    DOF *u;
    INT i, n = DofGetDataCount(u_h);
    FLOAT error = 0.0;

    for (i = 0; i < n; i++)
	if (Fabs(error = DofData(u_h)[i]) > 0.1)
	    break;
    error = (error > 0. ? phgDofNormL2(u_h) : - phgDofNormL2(u_h));
    u = phgDofNew(u_h->g, u_h->type, 1, "u", func_u);
    phgDofAXPY(-phgDofNormL2(u) / error, u_h, &u);
    error = phgDofNormL2(u);
    phgDofFree(&u);

    return error;
}

static void
build_matrices(MAT *matA, MAT *matB, DOF *u_h)
{
    int N = u_h->type->nbas;	/* number of basis functions in an element */
    int i, j;
    GRID *g = u_h->g;
    SIMPLEX *e;
    FLOAT A[N][N], B[N][N], buffer[N], vol;
    static FLOAT *B0 = NULL;
    INT I[N];

    if (B0 == NULL && g->nroot > 0) {
	/* (\int \phi_i\cdot\phi_j)/vol is independent of element */
	FreeAtExit(B0);
	B0 = phgAlloc(N * N * sizeof(*B0));
	e = g->roots;
	vol = phgGeomGetVolume(g, e);
	for (i = 0; i < N; i++)
	    for (j = 0; j <= i; j++)
		B0[i * N + j] = B0[i + j * N] =
		    phgQuadBasDotBas(e, u_h, i, u_h, j, QUAD_DEFAULT) / vol;
    }

    assert(u_h->dim == 1);
    ForAllElements(g, e) {
	vol = phgGeomGetVolume(g, e);
	for (i = 0; i < N; i++) {
	    I[i] = phgMapE2L(matA->cmap, 0, e, i);
	    for (j = 0; j <= i; j++) {
		/* \int \grad\phi_i\cdot\grad\phi_j */
		A[j][i] = A[i][j] =
		    phgQuadGradBasDotGradBas(e, u_h, i, u_h, j, QUAD_DEFAULT);
		/* \int \phi_i\cdot\phi_j */
		B[j][i] = B[i][j] = B0[i * N + j] * vol;
	    }
	}

	/* loop on basis functions */
	for (i = 0; i < N; i++) {
	    if (!phgDofDirichletBC(u_h, e, i, NULL, buffer, NULL,
					DOF_PROJ_NONE)) {
		phgMatAddEntries(matA, 1, I + i, N, I, A[i]); 
		phgMatAddEntries(matB, 1, I + i, N, I, B[i]); 
	    }
	}
    }
}

static FLOAT
estimate_error(FLOAT lambda, DOF *u_h, DOF *error)
/* compute H1 error indicator */
{
    GRID *g = u_h->g;
    SIMPLEX *e;
    DOF *grad_u, *jump, *residual, *tmp;

    grad_u = phgDofGradient(u_h, NULL, NULL, NULL);
    tmp = phgDofDivergence(grad_u, NULL, NULL, NULL);
    residual = phgDofCopy(u_h, NULL, DOF_DGn[u_h->type->order], NULL);
    phgDofAXPY(lambda, tmp, &residual);
    phgDofFree(&tmp);
    jump = phgQuadFaceJump(grad_u, DOF_PROJ_DOT, NULL, QUAD_DEFAULT);
    ForAllElements(g, e) {
	int i;
	FLOAT eta, h;
	FLOAT diam = phgGeomGetDiameter(g, e);
	eta = 0.0;
	/* for each face F compute [grad_u \cdot n] */
	for (i = 0; i < NFace; i++) {
	    if (e->bound_type[i] & (DIRICHLET | NEUMANN))
		continue;	/* boundary face */
	    h = phgGeomGetFaceDiameter(g, e, i);
	    eta += *DofFaceData(jump, e->faces[i]) * h;
	}
	eta = eta * .5 + diam * diam * phgQuadDofDotDof(e, residual, residual,
								-1);
	*DofElementData(error, e->index) = Sqrt(eta);
    }
    phgDofFree(&jump);
    phgDofFree(&residual);
    phgDofFree(&grad_u);
    return phgDofNormInftyVec(error);
}

static int
bc_map(int bctype)
{
    return DIRICHLET;	/* set Dirichlet BC on all boundaries */
}

#define MATRIX_FREE_TEST 0

#if MATRIX_FREE_TEST
/* callback function for matrix A */
static int
mat_vec(MAT_OP op, MAT *mat, VEC *x, VEC *y)
/* computes y = op(A) * x */
{
    phgMatVec(op, 1.0, ((MAT **)mat->mv_data)[0], x, 0.0, &y);
    return 0;
}
#endif	/* MATRIX_FREE_TEST */

int
main(int argc, char *argv[])
{
    static char *fn = "../test/cube.dat";
    /*static char *fn = "../test/model.mesh";*/
    size_t mem, mem_peak;
    int i, n, nit;
    INT nev = 10;
    INT pre_refines = 2;
    GRID *g;
    SIMPLEX *e;
    DOF **u_h, **error;
    MAP *map;
    MAT *A, *B;
    FLOAT tol = 1e-3, tau = 0.0, PEoo, thres;
    FLOAT *evals;
    FLOAT wtime;
    BOOLEAN show_analytic = FALSE;
#if MATRIX_FREE_TEST
    MAT *C;

    phgOptionsPreset("-arpack_solver pcg");
    phgOptionsPreset("-blopex_pc_type pcg");
#endif	/* MATRIX_FREE_TEST */

    phgOptionsRegisterFilename("mesh_file", "Mesh filename", &fn);
    phgOptionsRegisterInt("pre_refines", "Pre-refines", &pre_refines);
    phgOptionsRegisterInt("nev", "Number of eigenvalues", &nev);
    phgOptionsRegisterFloat("tol", "Convergence tolerance", &tol);
    phgOptionsRegisterFloat("tau", "The shift", &tau);
    phgOptionsRegisterNoArg("show_analytic", "Show analytic eigenvalues for "
			    "the unit cube", &show_analytic);

    phgInit(&argc, &argv);

    if (show_analytic) {
	if (phgRank == 0) {
	    for (n = 0, nit = 3; n < nev; nit++) {
		/* find i, j, k such that i^2 + j^2 + k^2 == nit */
		int j, k;
		for (i = 1; n < nev && i * i <= nit; i++) {
		    for (j = 1; n < nev && j * j <= nit - i * i; j++) {
			for (k = 1; k * k < nit - i * i - j * j; k++);
			if (i * i + j * j + k * k != nit)
			    continue;
			phgPrintf("tau[%d] = %0.16le\n", n++, nit*M_PI*M_PI);
		    }
		}
	    }
	}
	phgFinalize();
	exit(0);
    }

    g = phgNewGrid(-1);
    phgImportSetBdryMapFunc(bc_map);
    if (!phgImport(g, fn, FALSE))
	phgError(1, "can't read file \"%s\".\n", fn);
    /* pre-refinement */
    phgRefineAllElements(g, pre_refines);

    u_h = phgAlloc(nev * sizeof(*u_h));
    error = phgAlloc(nev * sizeof(*error));
    evals = phgCalloc(nev, sizeof(*evals));
    for (i = 0; i < nev; i++) {
	u_h[i] = phgDofNew(g, DOF_DEFAULT, 1, "u_h", DofInterpolation);
	phgDofSetDirichletBoundaryMask(u_h[i], BDRY_MASK);
	/* set random initial values for the eigenvectors */
	phgDofRandomize(u_h[i], i == 0 ? 123 : 0);
	error[i] = phgDofNew(g, DOF_P0, 1, "error indicator", DofNoAction);
    }

    while (TRUE) {
	phgPrintf("\n");
	if (phgBalanceGrid(g, 1.2, -1, NULL, 0.))
	    phgPrintf("Repartition mesh\n");
	phgPrintf("%d DOF, %d elements, %d submesh%s load imbalance: %lg\n",
			DofGetDataCountGlobal(u_h[0]), g->nleaf_global,
			g->nprocs, g->nprocs > 1 ? "es" : "", (double)g->lif);
	wtime = phgGetTime(NULL);
	map = phgMapCreate(u_h[0], NULL);
	A = phgMapCreateMat(map, map);
	B = phgMapCreateMat(map, map);
	build_matrices(A, B, u_h[0]);
	phgMatRemoveBoundaryEntries(A);
	phgMatRemoveBoundaryEntries(B);
	phgPrintf("Build matrices, wtime: %0.2lfs\n",
			(double)(phgGetTime(NULL) - wtime));
	wtime = phgGetTime(NULL);
#if MATRIX_FREE_TEST
	C = phgMapCreateMatrixFreeMat(A->rmap, A->cmap, mat_vec, A, NULL);
	n = phgMatEigenSolve(C, B, nev, EIGEN_CLOSEST, tau, &nit,
				evals, map, u_h, NULL);
	phgMatDestroy(&C);
#else	/* MATRIX_FREE_TEST */
	n = phgMatEigenSolve(A, B, nev, EIGEN_CLOSEST, tau, &nit,
				evals, map, u_h, NULL);
#endif	/* MATRIX_FREE_TEST */
	phgPrintf("%d iterations, converged eigenvalues: %d, wtime: %0.2lfs\n",
			nit, n, (double)(phgGetTime(NULL) - wtime));
	for (i = 0; i < n; i++)
	    phgPrintf("  tau[%d]=%0.6le, L2(u_h[%d])=%0.6le\n",
		i, (double)evals[i], i, (double)phgDofNormL2(u_h[i]));
	for (; i < nev; i++)
	    phgDofRandomize(u_h[i], 0);
	phgMatDestroy(&B);
	phgMatDestroy(&A);
	phgMapDestroy(&map);

	if (n == 0) {
	    phgRefineAllElements(g, 1);
	    continue;
	}

	PEoo = -1.0;
	for (i = 0; i < n; i++) {
	    thres = estimate_error(evals[i], u_h[i], error[i]);
	    if (PEoo < thres)
		PEoo = thres;
	}
	PEoo = Fabs(PEoo);
	phgPrintf("indicator=%0.3le, tau-tau_h=%0.3le, L2(u-u_h)=%0.3le\n",
		(double)PEoo, (double)(lambda0 - evals[0]),
		(double)compute_error(u_h[0]));
	mem = phgMemoryUsage(g, &mem_peak);
	phgPrintf("Memory usage: current %0.4lgMB, peak %0.4lgMB\n",
		(double)mem / (1024.0 * 1024.0),
		(double)mem_peak / (1024.0 * 1024.0));
	if (PEoo < tol || mem_peak >= 1024 * (size_t)1024 * 400)
	    break;
	/*thres = 1e-2 / Sqrt((double)g->nleaf_global);*/
	thres = PEoo * 0.5;
	ForAllElements(g, e)
	    e->mark = 0;
	for (i = 0; i < n; i++)
	    ForAllElements(g, e)
		if (*DofElementData(error[i],e->index) > thres)
		    e->mark = 1;
	phgRefineMarkedElements(g);
    }

#if 1
    phgPrintf("Creating \"%s\".\n", phgExportVTKn(g, "output.vtk", nev, u_h));
#endif

    phgFree(evals);
    for (i = 0; i < nev; i++) {
	phgDofFree(u_h + i);
	phgDofFree(error + i);
    }
    phgFree(u_h);
    phgFree(error);
    phgFreeGrid(&g);
    phgFinalize();
    return 0;
}
