/* Parallel Hierarchical Grid -- an adaptive finite element library.
 *
 * Copyright (C) 2005-2008 State Key Laboratory of Scientific and
 * Engineering Computing, Chinese Academy of Sciences. */

/* This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA */

/* $Id: map.c,v 1.83 2008/01/19 07:03:32 zlb Exp $ */

#define NEED_GetVariableArgs

/* The following macro controls whether DOFs in a vector are interleaved (1)
 * or concatenated (0).
 *
 * Warning: if INTERLEAVE_DOFS != 0 then the numbering of a multi-dof vector
 * is different when nprocs>1 then when nprocs==1. Also some applications
 * assume non interleaved ordering, so don't change its value unless for
 * testing purpose. */
#define INTERLEAVE_DOFS	0

#include "phg.h"

#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <stdarg.h>
#include <math.h>

/*------------ Functions managing maps ------------*/

#if USE_MPI
typedef struct {
    INT old_index;
    INT index;
    GTYPE type;
} INDEX;

static int
comp_index(const void *p0, const void *p1)
{
    const INDEX *pi0 = p0, *pi1 = p1;
    int i;

    if ((i = pi0->type - pi1->type) != 0)
	return i;
    return pi0->old_index - pi1->old_index;
}

#endif /* USE_MPI */

static INT *tmp_map = NULL;	/* local to global geometric indices map */

static INT
get_global_index(MAP *map, INT index)
/* maps local index 'index' to global index */
{
    GRID *g;
    DOF *u;
    INT m, n;
    INT index0;
    int dof_no, ndof = map->ndof;

    if (map->nprocs <= 1)
	return index;

    /* determine the dof_no of the index */
    for (dof_no = 0; dof_no < ndof; dof_no++) {
	if (index < map->offset[dof_no + 1])
	    break;
    }
    if (dof_no >= ndof) {
	phgError(1, "get_global_index: invalid local index: %d\n", index);
    }

    index -= map->offset[dof_no];
    u = map->dofs[dof_no];
    g = u->g;

    index0 = 0;

    if (u->count_vert > 0) {
	if (index < (n = (m = u->count_vert) * g->nvert))
	    return tmp_map[index / m] + index % m + map->count_vert[dof_no];
	index -= n;
    }
    index0 = (map->count_vert[ndof] > 0) ? g->nvert : 0;

    if (u->count_edge > 0) {
	if (index < (n = (m = u->count_edge) * g->nedge))
	    return tmp_map[index0 + index / m] + index % m
					       + map->count_edge[dof_no];
	index -= n;
    }
    index0 += (map->count_edge[ndof] > 0) ? g->nedge : 0;

    if (u->count_face > 0) {
	if (index < (n = (m = u->count_face) * g->nface))
	    return tmp_map[index0 + index / m] + index % m
					       + map->count_face[dof_no];
	index -= n;
    }
    index0 += (map->count_face[ndof] > 0) ? g->nface : 0;

    assert(u->count_elem > 0);
    m = u->count_elem;
    return tmp_map[index0 + index / m] + index % m + map->count_elem[dof_no];
}

#if USE_MPI
static INT
reorder(MAP *map, INT index, const INT owned_counts[], const INT offset[])
/* map an interleaved local index to corresponding concatenated local index.
 *	owned_counts[0] = the count of owned vertices,
 *	owned_counts[1] = the count of owned edges,
 *	owned_counts[2] = the count of owned faces,
 *	owned_counts[3] = the count of owned elements.
 *
 *	offset[i] is the starting local index of dof[i]
 */
{
    INT count, n;
    int i, no;

    if ((count = map->count_vert[map->ndof]) > 0) {
	n = count * owned_counts[0];
	if (index < n) {
	    no = phgBinarySearch(map->ndof, map->count_vert, i = index % count);
	    assert(no >= 0 && no < map->ndof);
	    index /= count;
	    i -= map->count_vert[no];
	    return offset[no] + index * map->dofs[no]->count_vert + i;
	}
	index -= n;
    }

    if ((count = map->count_edge[map->ndof]) > 0) {
	n = count * owned_counts[1];
	if (index < n) {
	    no = phgBinarySearch(map->ndof, map->count_edge, i = index % count);
	    assert(no >= 0 && no < map->ndof);
	    index /= count;
	    i -= map->count_edge[no];
	    return offset[no] + index * map->dofs[no]->count_edge + i +
			owned_counts[0] * map->dofs[no]->count_vert;
	}
	index -= n;
    }

    if ((count = map->count_face[map->ndof]) > 0) {
	n = count * owned_counts[2];
	if (index < n) {
	    no = phgBinarySearch(map->ndof, map->count_face, i = index % count);
	    assert(no >= 0 && no < map->ndof);
	    index /= count;
	    i -= map->count_face[no];
	    return offset[no] + index * map->dofs[no]->count_face + i +
			owned_counts[0] * map->dofs[no]->count_vert +
			owned_counts[1] * map->dofs[no]->count_edge;
	}
	index -= n;
    }

    count = map->count_elem[map->ndof];
    assert(index < count * owned_counts[3]);
    no = phgBinarySearch(map->ndof, map->count_elem, i = index % count);
    assert(no >= 0 && no < map->ndof);
    index /= count;
    i -= map->count_elem[no];
    return offset[no] + index * map->dofs[no]->count_elem + i +
		owned_counts[0] * map->dofs[no]->count_vert +
		owned_counts[1] * map->dofs[no]->count_edge +
		owned_counts[2] * map->dofs[no]->count_face;
}
#endif	/* USE_MPI */

static void
build_L2Vmap(MAP *map)
{
    int ii;
    DOF *u = map->dofs[0];
    GRID *g = u->g;
    INT i, j, k, m, n, index0;
    INT count_vert, count_edge, count_face, count_elem;
#if USE_MPI
    INT *counts = NULL, tmp[5];
    INT l, missing, available, max_missing;
    INT offset0, offset1, offset2, offset3, size;
    INT verts_owned = 0, edges_owned = 0, faces_owned = 0, elems_owned = 0;
    INDEX *aindex;		/* available list */
    INDEX *mindex;		/* missing list (to circulate) */
    int src, dst;
#endif

    ii = (int)map->ndof + 1;
    map->offset = phgAlloc(ii * sizeof(*map->offset));
    map->offset[0] = 0;
    for (i = 0; i < map->ndof; i++) {
	u = map->dofs[i];
	map->offset[i + 1] = map->offset[i] + DofGetDataCount(u);
    }

    if (map->count_vert == NULL) {
	map->count_vert = phgAlloc(ii * sizeof(*map->count_vert));
	map->count_edge = phgAlloc(ii * sizeof(*map->count_edge));
	map->count_face = phgAlloc(ii * sizeof(*map->count_face));
	map->count_elem = phgAlloc(ii * sizeof(*map->count_elem));

	count_vert = count_edge = count_face = count_elem = 0;
	for (ii = 0; ii < map->ndof; ii++) {
	    u = map->dofs[ii];
	    map->count_vert[ii] = count_vert;
	    map->count_edge[ii] = count_edge;
	    map->count_face[ii] = count_face;
	    map->count_elem[ii] = count_elem;
	    count_vert += u->count_vert;
	    count_edge += u->count_edge;
	    count_face += u->count_face;
	    count_elem += u->count_elem;
	}
	map->count_vert[ii] = count_vert;
	map->count_edge[ii] = count_edge;
	map->count_face[ii] = count_face;
	map->count_elem[ii] = count_elem;
    }
    else {
	count_vert = map->count_vert[ii];
	count_edge = map->count_edge[ii];
	count_face = map->count_face[ii];
	count_elem = map->count_elem[ii];
    }

    if (g->nprocs <= 1) {
	map->lif = 1.0;
	map->partition = phgAlloc(2 * sizeof(*map->partition));
	map->partition[0] = 0;
	map->partition[1] = map->nlocal;
	goto skip;
    }

#if USE_MPI
    offset0 = 0;
    offset1 = offset0 + (count_vert > 0 ? g->nvert : 0);
    offset2 = offset1 + (count_edge > 0 ? g->nedge : 0);
    offset3 = offset2 + (count_face > 0 ? g->nface : 0);
    size = offset3 + (count_elem > 0 ? g->nelem : 0);

    tmp_map = phgAlloc(size * sizeof(*tmp_map));
    map->partition = phgAlloc((g->nprocs + 1) * sizeof(*map->partition));
    map->partition[0] = 0;

    counts = phgAlloc(5 * g->nprocs * sizeof(*counts));

    missing = available = 0;

    /* TODO: implement an algorithm similar to pack_indices() in coarsen.c */
    if (count_vert > 0) {
	verts_owned = 0;
	for (i = 0; i < g->nvert; i++) {
	    if (g->types_vert[i] & OWNER) {
		verts_owned++;
		/*???if (g->types_vert[i] & REMOTE) */  {
		    available++;
		}
	    }
	    else {
		missing++;
	    }
	}
    }

    if (count_edge > 0) {
	edges_owned = 0;
	for (i = 0; i < g->nedge; i++) {
	    if (g->types_edge[i] & OWNER) {
		edges_owned++;
		/*???if (g->types_edge[i] & REMOTE) */  {
		    available++;
		}
	    }
	    else {
		missing++;
	    }
	}
    }

    if (count_face > 0) {
	faces_owned = 0;
	for (i = 0; i < g->nface; i++) {
	    if (g->types_face[i] & OWNER) {
		faces_owned++;
		/*???if (g->types_face[i] & REMOTE) */  {
		    available++;
		}
	    }
	    else {
		missing++;
	    }
	}
    }

    if (count_elem > 0) {
	elems_owned = 0;
	for (i = 0; i < g->nelem; i++) {
	    if (g->types_elem[i] & OWNER) {
		elems_owned++;
		/*???if (g->types_elem[i] & REMOTE) */  {
		    available++;
		}
	    }
	    else {
		missing++;
	    }
	}
	assert(elems_owned == g->nleaf);
    }

    /* gather all local counts */
    tmp[0] = missing;
    tmp[1] = verts_owned;
    tmp[2] = edges_owned;
    tmp[3] = faces_owned;
    tmp[4] = elems_owned;
    MPI_Allgather(tmp, 5 * sizeof(tmp[0]), MPI_BYTE,
		  counts, 5 * sizeof(*counts), MPI_BYTE, g->comm);

    max_missing = 0;
#if DEBUG
    verts_owned = edges_owned = faces_owned = elems_owned = 0;
#endif
    l = 0;			/* used to compute max nlocal */
    k = 0;			/* avoid gcc warning */
    for (i = 0; i < g->nprocs; i++) {
#if DEBUG
	verts_owned += counts[5 * i + 1];
	edges_owned += counts[5 * i + 2];
	faces_owned += counts[5 * i + 3];
	elems_owned += counts[5 * i + 4];
#endif
	/* compute nlocal for proc i */
	k = counts[5 * i + 1] * count_vert +
	    counts[5 * i + 2] * count_edge +
	    counts[5 * i + 3] * count_face + counts[5 * i + 4] * count_elem;
	if (l < k)
	    l = k;
	map->partition[i + 1] = map->partition[i] + k;
	j = counts[5 * i];
	if (max_missing < j)
	    max_missing = j;
    }
    map->lif = (l * (FLOAT)g->nprocs) / map->partition[g->nprocs];

    if (g->rank == 0) {
	/* test whether we have nlocal == 0 on some processes */
	j = 0;
	for (i = 0; i < g->nprocs; i++)
	    if (map->partition[i] == map->partition[i + 1])
		j++;
	if (j > 0) {
#if DEBUG
	    phgWarning("The map has local size 0 in the following process%s:\n",
		       j > 1 ? "es" : "");
	    for (i = 0; i < g->nprocs; i++)
		if (map->partition[i] == map->partition[i + 1])
		    phgWarning("\tproc %d (%s)\n", i, phgProcessorNames[i]);
	    phgWarning("which may confuse some solvers.\n");
#else	/* DEBUG */
	    phgWarning("the map has local size 0 in %d process%s, "
		       "which may confuse some solvers.\n",
		       j, j > 1 ? "es" : "");
#endif	/* DEBUG */
	}
    }

#if DEBUG
    if (count_vert > 0 && verts_owned != g->nvert_global)
	phgError(1, "%s:%d, something is wrong!\n", __FILE__, __LINE__);
    if (count_edge > 0 && edges_owned != g->nedge_global)
	phgError(1, "%s:%d, something is wrong!\n", __FILE__, __LINE__);
    if (count_face > 0 && faces_owned != g->nface_global)
	phgError(1, "%s:%d, something is wrong!\n", __FILE__, __LINE__);
    if (count_elem > 0 && elems_owned != g->nelem_global)
	phgError(1, "%s:%d, something is wrong!\n", __FILE__, __LINE__);
#endif /* DEBUG */

    /* assign new Global indices, collect available and missing entries */
    mindex = phgAlloc(2 * max_missing * sizeof(*mindex));
    aindex = phgAlloc(available * sizeof(*aindex));

    missing = available = 0;
    index0 = map->partition[g->rank];

    if (count_vert > 0) {
	for (i = 0; i < g->nvert; i++) {
	    if (g->types_vert[i] & OWNER) {
		tmp_map[offset0 + i] = index0;
		/*???if (g->types_vert[i] & REMOTE) */  {
		    aindex[available].type = VERTEX;
		    aindex[available].old_index = g->L2Gmap_vert[i];
		    aindex[available++].index = index0;
		}
		index0 += count_vert;
	    }
	    else {
		mindex[missing].type = VERTEX;
		mindex[missing].old_index = g->L2Gmap_vert[i];
		/* negate the index entry. It will be restored when old_index
		 * is changed to new gindex */
		mindex[missing++].index = -1 - (offset0 + i);
	    }
	}
    }

    if (count_edge > 0) {
	for (i = 0; i < g->nedge; i++) {
	    if (g->types_edge[i] & OWNER) {
		tmp_map[offset1 + i] = index0;
		/*???if (g->types_edge[i] & REMOTE) */  {
		    aindex[available].type = EDGE;
		    aindex[available].old_index = g->L2Gmap_edge[i];
		    aindex[available++].index = index0;
		}
		index0 += count_edge;
	    }
	    else {
		mindex[missing].type = EDGE;
		mindex[missing].old_index = g->L2Gmap_edge[i];
		/* negate the index entry. It will be restored when old_index
		 * is changed to new gindex */
		mindex[missing++].index = -1 - (offset1 + i);
	    }
	}
    }

    if (count_face > 0) {
	for (i = 0; i < g->nface; i++) {
	    if (g->types_face[i] & OWNER) {
		tmp_map[offset2 + i] = index0;
		/*???if (g->types_face[i] & REMOTE) */  {
		    aindex[available].type = FACE;
		    aindex[available].old_index = g->L2Gmap_face[i];
		    aindex[available++].index = index0;
		}
		index0 += count_face;
	    }
	    else {
		mindex[missing].type = FACE;
		mindex[missing].old_index = g->L2Gmap_face[i];
		/* negate the index entry. It will be restored when old_index
		 * is changed to new gindex */
		mindex[missing++].index = -1 - (offset2 + i);
	    }
	}
    }

    if (count_elem > 0) {
	for (i = 0; i < g->nelem; i++) {
	    if (g->types_elem[i] & OWNER) {
		tmp_map[offset3 + i] = index0;
		/*???if (g->types_elem[i] & REMOTE) */  {
		    aindex[available].type = ELEMENT;
		    aindex[available].old_index = g->L2Gmap_elem[i];
		    aindex[available++].index = index0;
		}
		index0 += count_elem;
	    }
	    else {
		mindex[missing].type = ELEMENT;
		mindex[missing].old_index = g->L2Gmap_elem[i];
		/* negate the index entry. It will be restored when old_index
		 * is changed to new gindex */
		mindex[missing++].index = -1 - (offset3 + i);
	    }
	}
    }

    /* sort lists with respect to (type, old gindex) */
    qsort(mindex, missing, sizeof(*mindex), comp_index);
    qsort(aindex, available, sizeof(*aindex), comp_index);

    /* circulate mindex[], picking up missing entries from foreign lists
     * FIXME: change to use ALltoallv for better efficiency */
    j = g->rank;
    k = 0;
    dst = (g->rank == 0) ? g->nprocs - 1 : g->rank - 1;
    src = (g->rank == g->nprocs - 1) ? 0 : g->rank + 1;
    while (TRUE) {
	MPI_Status status;
	if ((l = j + 1) >= g->nprocs)
	    l = 0;
	MPI_Sendrecv(mindex + k, sizeof(*mindex) * counts[5 * j], MPI_BYTE,
		     dst, 111, mindex + (max_missing - k),
		     sizeof(*mindex) * counts[5 * l], MPI_BYTE, src, 111,
		     g->comm, &status);
#if DEBUG
#if SENDRECV_TEST
	/* Verify result of MPI_Sendrecv by sending back and comparing data */
	{
	    INT i;
	    INDEX *back;
	    back = phgAlloc(sizeof(*mindex) * counts[5 * j]);
	    MPI_Sendrecv(mindex + (max_missing - k),
			 sizeof(*mindex) * counts[5 * l], MPI_BYTE, src, 222,
			 back, sizeof(*mindex) * counts[5 * j], MPI_BYTE, dst,
			 222, g->comm, &status);
	    for (i = 0; i < counts[5 * j]; i++) {
		if (phgMemcmp(back + i, mindex + k + i, sizeof(*back))) {
		    phgInfo(-1, "%s: sent to (%s): %s %d %d, got: %s %d %d\n",
			    phgProcessorNames[g->rank],
			    phgProcessorNames[dst],
			    GTypeName(mindex[k + i].type),
			    mindex[k + i].old_index,
			    mindex[k + i].index,
			    GTypeName(back[i].type), back[i].old_index,
			    back[i].index);
		    phgError(1, "%s:%d, testing MPI_Sendrecv failed!\n",
				__FILE__, __LINE__);
		}
	    }
	    phgFree(back);
	}
#endif	/* SENDRECV_TEST */
#endif /* DEBUG */
	k = max_missing - k;
	j = l;
	missing = k + counts[5 * j];
	if (j == g->rank)
	    break;
	/* pick up missing data */
	i = k;
	l = 0;
	while (l < available && i < missing) {
	    int c;
	    if (mindex[i].index >= 0) {
		i++;
		continue;
	    }
	    c = comp_index(aindex + l, mindex + i);
	    if (c < 0) {
		l++;
		continue;
	    }
	    else if (c > 0) {
		i++;
		continue;
	    }
	    mindex[i].old_index = aindex[l].index;
	    mindex[i].index = -1 - mindex[i].index;
	    l++;
	    i++;
	}
    }

    /* update missing entries */
    for (i = k; i < missing; i++) {
	if (mindex[i].index < 0) {
#if DEBUG
	    switch (mindex[i].type) {
		case VERTEX:
		    j = (-1 - mindex[i].index - offset0);
		    l = g->L2Gmap_vert[j];
		    ii = g->types_vert[j];
		    break;
		case EDGE:
		    j = (-1 - mindex[i].index - offset1);
		    l = g->L2Gmap_edge[j];
		    ii = g->types_edge[j];
		    break;
		case FACE:
		    j = (-1 - mindex[i].index - offset2);
		    l = g->L2Gmap_face[j];
		    ii = g->types_face[j];
		    break;
		case ELEMENT:
		    j = (-1 - mindex[i].index - offset3);
		    l = g->L2Gmap_elem[j];
		    ii = g->types_elem[j];
		    break;
	    }
	    if (ii != UNREFERENCED) {
		phgInfo(-1, "type=%s, index=%d (%d), flags=%d\n",
			GTypeName(mindex[i].type), j, l, ii);
		phgError(1, "%s:%d, something is wrong.\n", __FILE__,
			 __LINE__);
	    }
#endif /* DEBUG */
	}
	tmp_map[mindex[i].index] = mindex[i].old_index;
    }

    phgFree(aindex);
    phgFree(mindex);
#endif /* USE_MPI */

  skip:
    map->L2Vmap = phgAlloc(map->localsize * sizeof(*map->L2Vmap));
    map->ordering = NULL;
    if (map->localsize == map->nlocal)
	map->O2Gmap = NULL;
    else
	map->O2Gmap = phgAlloc((map->localsize - map->nlocal)
				  * sizeof(*map->O2Gmap));
    n = map->nlocal;
    i = 0;
    index0 = map->partition[map->rank];
    for (j = 0; j < map->ndof; j++) {
	u = map->dofs[j];
	m = DofGetDataCount(u);
	for (k = 0; k < m; k++, i++) {
	    INT l = get_global_index(map, i);
	    if (l >= index0 && l < index0 + map->nlocal) {
		map->L2Vmap[i] = l - index0;
	    }
	    else {
		map->O2Gmap[n - map->nlocal] = l;
		map->L2Vmap[i] = n++;
	    }
	}
    }
    assert(n == map->localsize);

    if (tmp_map != NULL)
	phgFree(tmp_map);

#if USE_MPI
#if !INTERLEAVE_DOFS
    if (g->nprocs > 1 && map->ndof > 1) {
	/* In the local vector, the entries of the DOFs are interleaved.
	 * New we reorder the global indices such that locally owned DOFs are
	 * concatenated in the local vector. */
	
	/* compute offsets in all local vectors of each DOF */
	tmp_map = phgAlloc(map->nprocs * map->ndof * sizeof(*tmp_map));
	for (i = 0; i < map->nprocs; i++) {
	    k = 0;
	    for (j = 0; j < map->ndof; j++) {
		tmp_map[i * map->ndof + j] = k;
		k += counts[i * 5 + 1] * map->dofs[j]->count_vert +
		     counts[i * 5 + 2] * map->dofs[j]->count_edge +
		     counts[i * 5 + 3] * map->dofs[j]->count_face +
		     counts[i * 5 + 4] * map->dofs[j]->count_elem;
	    }
	    assert(k == map->partition[i + 1] - map->partition[i]);
	}
	
	/* convert the indices */
	for (i = 0; i < map->localsize; i++) {
	    if ((j = map->L2Vmap[i]) < map->nlocal) {
		/* index owned by the local process */
		map->L2Vmap[i] = reorder(map, j, counts + 5 * map->rank + 1,
					 tmp_map + map->ndof * map->rank);
		assert(map->L2Vmap[i] < map->nlocal);
	    }
	    else {
		/* index owned by a remote process */
		n = map->O2Gmap[j -= map->nlocal];
		src = phgBinarySearch(map->nprocs, map->partition, n);
		/* note: src = the process owning this index */
		assert(src >=0 && src < map->nprocs && src != map->rank);
		k = map->partition[src];
		map->O2Gmap[j] = k + reorder(map, n - k, counts + 5 * src + 1,
					     tmp_map + map->ndof * src);
		assert(map->O2Gmap[j] >= map->partition[src] &&
		       map->O2Gmap[j] < map->partition[src + 1]);
	    }
	}
	phgFree(tmp_map);
    }
#endif	/* !INTERLEAVE_DOFS */
    phgFree(counts);
#endif	/* USE_MPI */

    return;
}

INT
phgMapL2G(MAP *map, INT index)
{
    DOF *u = map->dofs[0];
    INT i;

    assert(index >= 0 && index < map->localsize);

    if (u->g->serial_no != map->serial_no)
	phgError(1, "phgMapL2G: unexpected grid change, abort.\n");

    if (u->g->nprocs == 1)
	return index;

    i = (map->L2Vmap == NULL ? index : map->L2Vmap[index]);
    assert(i >= 0 && i < map->localsize);

    if (i < map->nlocal)
	return map->partition[map->rank] + i;
    else
	return map->O2Gmap[i - map->nlocal];
}

static INT
get_local_size(MAP *map)
{
    GRID *g = (map->dofs == NULL ? NULL : map->dofs[0]->g);
    DOF *u;
    INT i, n, size = 0;
    int j;

    for (j = 0; j < map->ndof; j++) {
	u = map->dofs[j];

	if ((n = u->count_vert) > 0)
	    for (i = 0; i < g->nvert; i++) {
		if (g->types_vert[i] & OWNER)
		    size += n;
	    }

	if ((n = u->count_edge) > 0)
	    for (i = 0; i < g->nedge; i++) {
		if (g->types_edge[i] & OWNER)
		    size += n;
	    }

	if ((n = u->count_face) > 0)
	    for (i = 0; i < g->nface; i++) {
		if (g->types_face[i] & OWNER)
		    size += n;
	    }

	size += u->count_elem * g->nleaf;
    }

    return size;
}

static void
get_bdry_map(MAP *map)
/* initialize map->bdry_* members */
{
    GRID *g;
    DOF *dof;
    int ii, n;
    INT *counts, count;
    INT i, j, k, index, index0, gindex0;

    BTYPE b, *tmp_types[4];
    int tmp_count[4];
    INT tmp_size[4];
    BTYPE DB_mask;
#if USE_MPI
    int rank, *scnts, *sdsps, *rcnts, *rdsps;
    INT n0, ssize, rsize, *sbuf, *rbuf;
#endif	/* USE_MPI */

    if (map->dofs == NULL)
	return;

    g = map->dofs[0]->g;

    for (i = 0; i < map->ndof; i++)
        if (map->dofs[i]->DB_mask != 0)
	    break;
    if (i >= map->ndof)
	return;

    gindex0 = map->partition[g->rank];

    tmp_types[0] = g->types_vert;
    tmp_types[1] = g->types_edge;
    tmp_types[2] = g->types_face;
    tmp_types[3] = g->types_elem;

    tmp_size[0] = g->nvert;
    tmp_size[1] = g->nedge;
    tmp_size[2] = g->nface;
    tmp_size[3] = g->nelem;

    /* count number of boundary DOF owned by the present process.
     * Note: by putting the loop on map->dofs as the inner most one,
     * then entries in map->bdry[] are increasing */
    count = 0;
    phgFree(map->bdry);
    map->bdry = NULL;
    map->bdry_nlocal = 0;
    for (i = 0; i < map->ndof; i++) {
	dof = map->dofs[i];
	DB_mask = dof->DB_mask;
	if (DB_mask == 0)
	    continue;
	tmp_count[0] = dof->count_vert;
	tmp_count[1] = dof->count_edge;
	tmp_count[2] = dof->count_face;
	tmp_count[3] = dof->count_elem;
	index0 = 0;
	for (ii = 0; ii < 4; ii++) {
	    if ((n = tmp_count[ii]) == 0)
		continue;
	    for (j = 0; j < tmp_size[ii]; j++) {
		if (!((b = tmp_types[ii][j]) & DB_mask))
		    continue;
		if (count + n > map->bdry_nlocal) {
		    map->bdry_nlocal = (count + n + 1024);
		    map->bdry = phgRealloc(map->bdry, map->bdry_nlocal *
							sizeof(*map->bdry));
		}
		for (index = index0 + n * j;
		     index < index0 + n * (j + 1); index++) {
		    k = map->L2Vmap[phgMapD2L(map, i, index)];
		    assert(k < map->localsize);
#if 0
		    /* only mark local bdry entries */
		    if (k >= map->nlocal)
			continue;
#endif
		    map->bdry[count++] = k;
		}
	    }
	    index0 += tmp_size[ii] * n;
	}
    }

    if (count > 0) {
	/* sort entries in bdry[] */
	qsort(map->bdry, count, sizeof(*map->bdry), phgCompINT);
    }

    map->bdry_localsize = count;
    for (i = 0; i < count && map->bdry[i] < map->nlocal; i++);
    map->bdry_nlocal = count = i;

    /* gather counts from all processes */
    counts = phgAlloc(g->nprocs * sizeof(*counts));
#if USE_MPI
    MPI_Allgather(&count, 1, PHG_MPI_INT, counts, 1, PHG_MPI_INT, g->comm);
    k = 0;
    for (i = 0; i < g->nprocs; i++) {
	map->bdry_partition[i] = k;
	k += counts[i];
    }
    map->bdry_partition[i] = k;
#else /* USE_MPI */
    k = counts[0] = count;
    map->bdry_partition[0] = 0;
    map->bdry_partition[1] = count;
#endif /* USE_MPI */
    map->bdry_nglobal = k;

    phgInfo(2, "nlocal: %d, boundary DOF: %d (%d total)\n",
	    map->nlocal, count, k);

    phgFree(counts);

#if USE_MPI
    /* some unreferenced bdry entries may be missing, the next step is to
     * find them from other processes. */

    if (map->bdry_nglobal == 0)
	return;

    rsize = map->localsize - map->nlocal;

    /* count off-proc columns */
    scnts = phgAlloc(4 * g->nprocs * sizeof(*scnts));
    sdsps = scnts + map->nprocs;
    rcnts = sdsps + map->nprocs;
    rdsps = rcnts + map->nprocs;
    rbuf = phgAlloc(rsize * sizeof(*rbuf));
    rank = 0;
    bzero(rcnts, g->nprocs * sizeof(*rcnts));
    for (i = 0; i < map->localsize - map->nlocal; i++) {
	rbuf[i] = k = map->O2Gmap[map->ordering[i]];
	while (rank < g->nprocs && map->partition[rank + 1] <= k)
	    rank++;
	if (rank >= g->nprocs || rank == g->rank)
	    phgError(-1, "%s:%d, unexpected.\n", __FILE__, __LINE__);
	rcnts[rank]++;
    }
    MPI_Alltoall(rcnts, 1, MPI_INT, scnts, 1, MPI_INT, g->comm);
    i = j = 0;
    for (rank = 0; rank < g->nprocs; rank++) {
	rdsps[rank] = i;
	sdsps[rank] = j;
	i += rcnts[rank];
	j += scnts[rank];
    }
    sbuf = phgAlloc((ssize = j) * sizeof(*sbuf));
    phgAlltoallv(rbuf, rcnts, rdsps, PHG_MPI_INT,
		 sbuf, scnts, sdsps, PHG_MPI_INT, g->comm);
    n0 = map->partition[g->rank];
    if (ssize > 0 && map->bdry_nlocal > 0) {
	for (i = 0; i < ssize; i++) {
	    k = sbuf[i] - n0;
	    assert(k >= 0 && k < map->nlocal);
	    if (bsearch(&k, map->bdry, map->bdry_nlocal,
			sizeof(k), phgCompINT) != NULL)
		sbuf[i] = -1;	/* boundary entry */
	}
    }
    phgAlltoallv(sbuf, scnts, sdsps, PHG_MPI_INT,
		 rbuf, rcnts, rdsps, PHG_MPI_INT, g->comm);
    for (i = 0, k = 0; i < rsize; i++) {
	if (rbuf[i] < 0)
	    k++;
#if DEBUG
	else {
	    INT key = map->ordering[i] + map->nlocal;
	    assert((j = map->bdry_localsize - map->bdry_nlocal) == 0 ||
		   bsearch(&key, map->bdry + map->bdry_nlocal, j, sizeof(INT),
			   	phgCompINT) == NULL);
	}
#endif	/* DEBUG */
    }
    assert(k >= map->bdry_localsize - map->bdry_nlocal);
    if (k > 0) {
	k += map->bdry_nlocal;
	map->bdry = phgRealloc(map->bdry, k * sizeof(*map->bdry));
	for (i = 0, k = 0; i < rsize; i++) {
	    if (rbuf[i] >= 0)
		continue;
	    map->bdry[map->bdry_nlocal + k] = map->nlocal + map->ordering[i];
	    k++;
	}
	qsort(map->bdry + map->bdry_nlocal, k, sizeof(k), phgCompINT);
	map->bdry_localsize = map->bdry_nlocal + k;
    }
    phgFree(scnts);
    phgFree(rbuf);
    phgFree(sbuf);
#endif	/* USE_MPI */

#if 0
    k = map->partition[map->rank];
    phgInfo(-1, "n0 = %d, nlocal = %d, localsize = %d, nglobal = %d\n",
		    k, map->nlocal, map->localsize, map->nglobal);
    phgInfo(-1, "bdry_nlocal = %d\n", map->bdry_nlocal);
    for (i = 0; i < map->bdry_nlocal; i++) {
	assert(map->bdry[i] < map->nlocal);
	assert(i == 0 || map->bdry[i] > map->bdry[i - 1]);
	phgInfo(-1, "bdry[%d] = %d\n", i, k + map->bdry[i]);
    }
    phgInfo(-1, "bdry_localsize = %d\n", map->bdry_localsize);
    for (; i < map->bdry_localsize; i++) {
	k = map->bdry[i] - map->nlocal;
	assert(i == map->bdry_nlocal || map->bdry[i] > map->bdry[i - 1]);
	assert(k >= 0);
	phgInfo(-1, "bdry[%d] = %d (%d)\n", i, map->O2Gmap[k], k);
    }
    phgInfo(-1, "\n");
    if (map->localsize == map->nlocal)
	return;
    j = map->O2Gmap[map->ordering[0]];
    for (i = 1; i < map->localsize - map->nlocal; i++) {
	k = map->O2Gmap[map->ordering[i]];
	assert(k > j);
	j = k;
    }
#endif
}

MAP *
phgMapCreateSimpleMap(GRID *g, INT n, INT N)
{
    MAP *map = phgCalloc(1, sizeof(*map));
    int i;

    assert(n != PHG_DECIDE || N != PHG_DECIDE);

#if USE_MPI
    map->comm = g->comm;
#endif	/* USE_MPI */
    map->nprocs = g->nprocs;
    map->rank = g->rank;
    map->serial_no = g->serial_no;		/* save serial number */

    if (n == PHG_DECIDE) {
	n = N / map->nprocs;
	if (map->rank < N % map->nprocs)
	    n++;
    }
    assert(N == PHG_DECIDE || n <= N);
    map->nlocal = n;
    map->partition = phgAlloc((map->nprocs + 1) * sizeof(*map->partition));
    map->partition[0] = 0;
#if USE_MPI
    MPI_Allgather(&n, 1, PHG_MPI_INT, map->partition + 1, 1, PHG_MPI_INT,
		  map->comm);
#else	/* USE_MPI */
    map->partition[1] = n;
#endif	/* USE_MPI */
    for (i = 0; i < map->nprocs; i++)
	map->partition[i+ 1] += map->partition[i];
    map->nglobal = map->partition[i];
    if (N == PHG_DECIDE)
	N = map->nglobal;

    if (map->nglobal > N) {
	/* special case: identical serial map on all processes */
	assert(map->nglobal == map->nprocs * N);
	map->nglobal = N;
#if USE_MPI
	map->comm = MPI_COMM_SELF;
#endif	/* USE_MPI */
	map->nprocs = 1;
	map->rank = 0;
	phgFree(map->partition);
	map->partition = phgAlloc(2 * sizeof(*map->partition));
	map->partition[0] = 0;
	map->partition[1] = N;
    }

    assert(map->nglobal == N);
    map->localsize = map->nlocal;

    map->bdry_nlocal = map->bdry_nglobal = map->bdry_localsize = 0;
    map->bdry = NULL;
    map->bdry_partition =
	phgCalloc(map->nprocs + 1, sizeof(*map->bdry_partition));

    return map;
}

#if USE_MPI
static INT *O2Gmap;

static int
O2Gmap_comp(const void *p0, const void *p1)
{
    return O2Gmap[*(const INT *)p0] - O2Gmap[*(const INT *)p1];
}

void
_phg_build_ordering(size_t size, INT map[], INT **ordering)
/* build array ordering[] such that map[ordering[i]] is increasing w.r.t. i */
{
    INT i;

    if (size <= 0) {
	*ordering = NULL;
	return;
    }
    if (*ordering != NULL)
	return;

    O2Gmap = map;
    *ordering = phgAlloc(size * sizeof(**ordering));
    for (i = 0; i < size; i++)
	(*ordering)[i] = i;
    qsort(*ordering, size, sizeof(**ordering), O2Gmap_comp);
}
#endif /* USE_MPI */

MAP *
phgMapCreateN(int ndof, DOF **dofs)
{
    MAP *map = phgCalloc(1, sizeof(*map));
    GRID *g;
    DOF *u;
    int j;

    assert(dofs != NULL);

    map->ndof = ndof;
    map->dofs = phgAlloc(ndof * sizeof(*map->dofs));
    memcpy(map->dofs, dofs, ndof * sizeof(*map->dofs));
    u = dofs[0];
    g = u->g;
#if USE_MPI
    map->comm = g->comm;
#endif	/* USE_MPI */
    map->nprocs = g->nprocs;
    map->rank = g->rank;
    map->serial_no = g->serial_no;		/* save serial number */

    map->nlocal = get_local_size(map);
    map->nglobal = map->localsize = 0;
    for (j = 0; j < map->ndof; j++) {
	u = map->dofs[j];
	map->nglobal += DofGetDataCountGlobal(u);
	map->localsize += DofGetDataCount(u);
    }

    build_L2Vmap(map);

#if USE_MPI
    _phg_build_ordering(map->localsize - map->nlocal, map->O2Gmap,
							&map->ordering);
#endif	/* USE_MPI */

    map->bdry_nlocal = map->bdry_nglobal = map->bdry_localsize = 0;
    map->bdry = NULL;
    map->bdry_partition =
	phgCalloc(g->nprocs + 1, sizeof(*map->bdry_partition));
    get_bdry_map(map);

    return map;
}

MAP *
phgMapCreate(DOF *u, ...)
{
    int ndof;
    DOF **dofs;
    MAP *map;

    GetVariableArgs(ndof, dofs, u, DOF *);
    map = phgMapCreateN(ndof, dofs);
    phgFree(dofs);

    return map;
}

void
phgMapDestroy(MAP **map_ptr)
{
    MAP *map = *map_ptr;

    if (map == NULL)
	return;

    *map_ptr = NULL;

    phgMapDestroy(&map->map_nb);

    assert(map->ref_count >= 0);

    if (map->ref_count > 0) {
	map->ref_count--;
	return;
    }

    phgFree(map->L2Vmap);
    phgFree(map->O2Gmap);
    phgFree(map->ordering);
    phgFree(map->partition);
    phgFree(map->offset);
    phgFree(map->bdry);
    phgFree(map->bdry_partition);

    phgFree(map->count_vert);
    phgFree(map->count_edge);
    phgFree(map->count_face);
    phgFree(map->count_elem);

    phgFree(map->dofs);

#if USE_MPI
    phgMapDestroyCommInfo(&map->cinfo);
#endif	/* USE_MPI */

    phgFree(map);

    return;
}

MAP *
phgMapRemoveBoundaryEntries(MAP *map)
/* derives a new map from 'map' with all boundary entries removed */
{
    INT n, N;

    if (map->bdry_nglobal == 0) {
	map->ref_count++;
	return map;
    }

    if (map->map_nb != NULL) {
	map->map_nb->ref_count++;
	return map->map_nb;
    }
 
    n = map->nlocal - map->bdry_nlocal;
    N = map->nglobal - map->bdry_nglobal;
    assert(map->dofs != NULL);
    map->map_nb = phgMapCreateSimpleMap(map->dofs[0]->g, n, N);
    map->map_nb->ref_count++;	/* note: map_nb is referenced by map */

    return map->map_nb;
}

/*--------- functions for vector scattering/gathering ---------*/

#if USE_MPI
COMM_INFO *
phgMapCreateCommInfo(MPI_Comm comm, int nprocs, int rank, INT nlocal,
		     INT localsize, INT *O2Gmap, INT *ordering, INT *partition)
/* Note: ordering may be NULL which means the natural ordering */
{
    COMM_INFO *cinfo;
    INT i, k, n, *work, O2Gsize = localsize - nlocal;
    int ii, mask;

    if (nprocs == 1)
	return NULL;

    cinfo = phgCalloc(1, sizeof(*cinfo));
    cinfo->comm = comm;

    cinfo->scnts = phgAlloc(4 * nprocs * sizeof(*cinfo->scnts));
    cinfo->sdsps = cinfo->scnts + nprocs;
    cinfo->rcnts = cinfo->sdsps + nprocs;
    cinfo->rdsps = cinfo->rcnts + nprocs;

    ii = 0;
    bzero(cinfo->rcnts, nprocs * sizeof(*cinfo->rcnts));
    mask = 0;
    if (ordering != NULL) {
	work = phgAlloc(O2Gsize * sizeof(*work));
	for (i = 0; i < O2Gsize; i++) {
	    k = work[i] = O2Gmap[ordering[i]];
	    while (ii < nprocs && partition[ii + 1] <= k)
		ii++;
	    assert(ii < nprocs && ii != rank);
	    cinfo->rcnts[ii]++;
	    mask = 1;
	}
    }
    else {
	work = O2Gmap;
	for (i = 0; i < O2Gsize; i++) {
	    k = O2Gmap[i];
	    while (ii < nprocs && partition[ii + 1] <= k)
		ii++;
	    assert(ii < nprocs && ii != rank);
	    cinfo->rcnts[ii]++;
	    mask = 1;
	}
    }

    /* attach mask to rcnts[] */
    if (mask != 0) {
	mask = (mask << (sizeof(mask) * 8 - 1));
	for (ii = 0; ii < nprocs; ii++)
	    cinfo->rcnts[ii] |= mask;
    }

    MPI_Alltoall(cinfo->rcnts, 1, MPI_INT, cinfo->scnts, 1, MPI_INT, comm);

    /* check and remove mask */
    mask = (1 << (sizeof(mask) * 8 - 1));
    n = 0;
    for (ii = 0; ii < nprocs; ii++) {
	if ((cinfo->scnts[ii] & mask))
	    n = 1;
	cinfo->scnts[ii] &= ~mask;
	cinfo->rcnts[ii] &= ~mask;
    }
    if (n == 0) {
	/* no communication needed for this map */
	assert(work == NULL);
	phgFree(cinfo->scnts);
	cinfo->scnts = cinfo->sdsps = cinfo->rcnts = cinfo->rdsps = NULL;
	cinfo->comm = MPI_COMM_NULL;
	return cinfo;
    }

    cinfo->size = nlocal;
    cinfo->rind = ordering;

    cinfo->rsize = cinfo->ssize = 0;
    for (ii = 0; ii < nprocs; ii++) {
	cinfo->rdsps[ii] = cinfo->rsize;
	cinfo->sdsps[ii] = cinfo->ssize;
	cinfo->rsize += cinfo->rcnts[ii];
	cinfo->ssize += cinfo->scnts[ii];
    }
    assert(cinfo->rsize == O2Gsize);
    /* prepare indices of data to send */
    cinfo->sind = phgAlloc(cinfo->ssize * sizeof(*cinfo->sind));
    phgAlltoallv(work,	      cinfo->rcnts, cinfo->rdsps, PHG_MPI_INT,
		 cinfo->sind, cinfo->scnts, cinfo->sdsps, PHG_MPI_INT, comm);
    if (work != O2Gmap)
	phgFree(work);
    /* convert indices in sind[] to local */
    n = partition[rank];
    for (i = 0; i < cinfo->ssize; i++) {
	assert(n <= cinfo->sind[i] && cinfo->sind[i] < partition[rank + 1]);
	cinfo->sind[i] -= n;
    }

    assert(cinfo->rcnts[rank] == 0 && cinfo->scnts[rank] == 0);

    return cinfo;
}

void
phgMapDestroyCommInfo(COMM_INFO **cinfo)
{
    if (*cinfo == NULL)
	return;

    assert((*cinfo)->nsends == 0 && (*cinfo)->nrecvs == 0
		&& (*cinfo)->buffer == NULL);

    phgFree((*cinfo)->scnts);
    phgFree((*cinfo)->sind);
    phgFree(*cinfo);

    *cinfo = NULL;
}

void
phgMapGatherBegin(COMM_INFO *cinfo, int nvec, FLOAT *data, FLOAT *offp_data)
{
    int ii;
    INT i;
    const INT *rind;
    FLOAT *sbuff, *rbuff, *p, *q;
    MPI_Datatype type;

    if (nvec == 0 || cinfo == NULL || cinfo->comm == MPI_COMM_NULL)
	return;

    assert(cinfo->nsends == 0 && cinfo->nrecvs == 0 && cinfo->buffer == NULL);

    if (cinfo->ssize == 0 && cinfo->rsize == 0)
	return;

    if (cinfo->rind != NULL || nvec > 1) {
	/* Note: need rbuff to reorder the data if nvec > 1
	 * (FIXME: use MPI_Datatype with ex=1, size=rsize?) */
	cinfo->buffer = phgAlloc(nvec * (cinfo->ssize + cinfo->rsize)
				      * sizeof(*cinfo->buffer));
	sbuff = cinfo->buffer;
	rbuff = sbuff + nvec * cinfo->ssize;
    }
    else {
	cinfo->buffer = phgAlloc(nvec * cinfo->ssize * sizeof(*cinfo->buffer));
	sbuff = cinfo->buffer;
	rbuff = offp_data;
    }

    if (nvec > 1) {
	MPI_Type_contiguous(nvec, PHG_MPI_FLOAT, &type);
	MPI_Type_commit(&type);
    }
    else {
	type = PHG_MPI_FLOAT;
    }

    /* copy data to send from offp_data to rbuff */
    if (cinfo->rind != NULL) {
	q = rbuff;
	rind = cinfo->rind;
	if (nvec == 1) {
	    /* faster code for nvec == 1 */
	    for (i = 0; i < cinfo->rsize; i++)
		*(q++) = offp_data[*(rind++)];
	}
	else {
	    for (i = 0; i < cinfo->rsize; i++) {
		p = offp_data + *(rind++);
		for (ii = 0; ii < nvec; ii++) {
		    *(q++) = *p;
		    p += cinfo->rsize;
		}
	    }
	}
    }
    else if (nvec > 1) {	/* no need to copy data if nvec == 1 */
	q = rbuff;
	for (i = 0; i < cinfo->rsize; i++) {
	    p = offp_data++;
	    for (ii = 0; ii < nvec; ii++) {
		*(q++) = *p;
		p += cinfo->rsize;
	    }
	}
    }

    phgAlltoallvBegin(rbuff, cinfo->rcnts, cinfo->rdsps, type,
		       sbuff, cinfo->scnts, cinfo->sdsps, type, cinfo->comm,
		       &cinfo->requests, &cinfo->nsends, &cinfo->nrecvs);

    if (nvec > 1)
	MPI_Type_free(&type);
}

void
phgMapGatherEnd(COMM_INFO *cinfo, int nvec, FLOAT *data, FLOAT *offp_data)
{
    int ii;
    INT i, *sind;
    FLOAT *p, *q;

    if (nvec == 0 || cinfo == NULL || cinfo->comm == MPI_COMM_NULL)
	return;

    if (cinfo->nsends == 0 && cinfo->nrecvs == 0)
	return;

    phgAlltoallvEnd(cinfo->nsends, cinfo->nrecvs, &cinfo->requests);
    cinfo->nsends = cinfo->nrecvs = 0;

    p = cinfo->buffer;
    sind = cinfo->sind;
    if (nvec == 1) {
	/* faster code for nvec == 1 */
	for (i = 0; i < cinfo->ssize; i++) {
	    data[*(sind++)] += *(p++);
	}
    }
    else {
	for (i = 0; i < cinfo->ssize; i++) {
	    q = data + *(sind++);
	    for (ii = 0; ii < nvec; ii++) {
		*q += *(p++);
		q += cinfo->size;
	    }
	}
    }

    phgFree(cinfo->buffer);
    cinfo->buffer = NULL;
}

void
phgMapGather(COMM_INFO *cinfo, int nvec, FLOAT *data, FLOAT *offp_data)
/* data += offp_data */
{
    phgMapGatherBegin(cinfo, nvec, data, offp_data);
    phgMapGatherEnd(cinfo, nvec, data, offp_data);
}

void
phgMapScatterBegin(COMM_INFO *cinfo, int nvec, FLOAT *data, FLOAT *offp_data)
{
    int ii;
    INT i, *sind;
    FLOAT *sbuff, *rbuff, *p, *q;
    MPI_Datatype type;

    if (nvec == 0 || cinfo == NULL || cinfo->comm == MPI_COMM_NULL)
	return;

    assert(cinfo->nsends == 0 && cinfo->nrecvs == 0 && cinfo->buffer == NULL);

    if (cinfo->ssize == 0 && cinfo->rsize == 0)
	return;

    if (nvec > 1) {
	static BOOLEAN warned = FALSE;
	if (!warned) {
	    phgWarning("function %s has not been tested!\n", __func__);
	    warned = TRUE;
	}
    }

    if (cinfo->rind != NULL || nvec > 1) {
	/* Note: need rbuff to reorder the data if nvec > 1
	 * (FIXME: use MPI_Datatype with ex=1, size=rsize?) */
	cinfo->buffer = phgAlloc(nvec * (cinfo->ssize + cinfo->rsize)
				      * sizeof(*cinfo->buffer));
	sbuff = cinfo->buffer;
	rbuff = sbuff + nvec * cinfo->ssize;
    }
    else {
	cinfo->buffer = phgAlloc(nvec * cinfo->ssize * sizeof(*cinfo->buffer));
	sbuff = cinfo->buffer;
	rbuff = offp_data;
    }

    if (nvec > 1) {
	MPI_Type_contiguous(nvec, PHG_MPI_FLOAT, &type);
	MPI_Type_commit(&type);
    }
    else {
	type = PHG_MPI_FLOAT;
    }

    /* copy data to send from data to sbuff */
    p = sbuff;
    sind = cinfo->sind;
    if (nvec == 1) {
	/* faster code for nvec == 1 */
	for (i = 0; i < cinfo->ssize; i++) {
	    *(p++) = data[*(sind++)];
	}
    }
    else {
	for (i = 0; i < cinfo->ssize; i++) {
	    q = data + *(sind++);
	    for (ii = 0; ii < nvec; ii++) {
		*(p++) = *q;
		q += cinfo->size;
	    }
	}
    }

    phgAlltoallvBegin(sbuff, cinfo->scnts, cinfo->sdsps, type,
		       rbuff, cinfo->rcnts, cinfo->rdsps, type, cinfo->comm,
		       &cinfo->requests, &cinfo->nsends, &cinfo->nrecvs);

    if (nvec > 1)
	MPI_Type_free(&type);
}

void
phgMapScatterEnd(COMM_INFO *cinfo, int nvec, FLOAT *data, FLOAT *offp_data)
{
    int ii;
    INT i;
    const INT *rind;
    FLOAT *p, *q;

    if (nvec == 0 || cinfo == NULL || cinfo->comm == MPI_COMM_NULL)
	return;

    if (cinfo->nsends == 0 && cinfo->nrecvs == 0)
	return;

    phgAlltoallvEnd(cinfo->nsends, cinfo->nrecvs, &cinfo->requests);
    cinfo->nsends = cinfo->nrecvs = 0;

    if (cinfo->rind != NULL) {
	q = cinfo->buffer + nvec * cinfo->ssize;
	rind = cinfo->rind;
	if (nvec == 1) {
	    /* faster code for nvec == 1 */
	    for (i = 0; i < cinfo->rsize; i++) {
		offp_data[*(rind++)] = *(q++);
	    }
	}
	else {
	    for (i = 0; i < cinfo->rsize; i++) {
		p = offp_data + *(rind++);
		for (ii = 0; ii < nvec; ii++) {
		    *p = *(q++);
		    p += cinfo->rsize;
		}
	    }
	}
    }
    else if (nvec > 1) {	/* data already in offp_data if nvec == 1 */
	q = cinfo->buffer + nvec * cinfo->ssize;
	for (i = 0; i < cinfo->rsize; i++) {
	    p = offp_data++;
	    for (ii = 0; ii < nvec; ii++) {
		*p = *(q++);
		p += cinfo->rsize;
	    }
	}
    }

    phgFree(cinfo->buffer);
    cinfo->buffer = NULL;
}

void
phgMapScatter(COMM_INFO *cinfo, int nvec, FLOAT *data, FLOAT *offp_data)
/* offp_data := data */
{
    phgMapScatterBegin(cinfo, nvec, data, offp_data);
    phgMapScatterEnd(cinfo, nvec, data, offp_data);
}
#endif	/* USE_MPI */

/*--------- functions for transferring data between DOF list and VEC --------*/

int
phgMapLocalDataToDof(MAP *map, int ndof, DOF **dofs, FLOAT *data)
/* This function copies the solution back to DOFs.
 * 'vec' is the pointer to the local vector data (size = map->nlocal)
 */
{
#if 1	/*==================================================================*/
  /* The old circular algorithm */
  {
    DOF *u;
    INT i, j, k = 0, l, m, nmax;
    INT nlocal = map->nlocal;
    /* struct for storing remote data */
    struct {
	FLOAT value;
	INT gindex;
    } *rdata = NULL;
    INT *offp_counts = NULL;
#if USE_MPI
    MPI_Status status;
    MPI_Datatype type;
    INT first, last;
    int src, dst;

    if (map->nprocs > 1) {
	offp_counts = phgAlloc(map->nprocs * sizeof(*offp_counts));
	m = map->localsize - nlocal;
	MPI_Allgather(&m, 1, PHG_MPI_INT, offp_counts, 1, PHG_MPI_INT,
		      map->comm);
    }
#endif

    nmax = 0;
    if (offp_counts != NULL) {
	for (i = 0; i < map->nprocs; i++) {
	    if (nmax < offp_counts[i])
		nmax = offp_counts[i];
	}
    }
    if (nmax == 0) {
	k = 0;
	goto skip_comm;
    }

#if USE_MPI
    /* TODO: better use Alltoallv */

    /* copy indices of missing data */
    rdata = phgAlloc(2 * nmax * sizeof(*rdata));
    m = map->localsize - map->nlocal;
    for (i = 0; i < m; i++)
	rdata[i].gindex = map->O2Gmap[i];

    /* circulate lists of missing data and collecting missing data */
    j = map->rank;
    k = 0;
    dst = (map->rank == 0) ? map->nprocs - 1 : map->rank - 1;
    src = (map->rank == map->nprocs - 1) ? 0 : map->rank + 1;
    MPI_Type_contiguous(sizeof(*rdata), MPI_BYTE, &type);
    MPI_Type_commit(&type);
    first = map->partition[map->rank];
    last = first + nlocal;
    for (i = 0; i < map->nprocs; i++) {
	if ((l = j + 1) >= map->nprocs)
	    l = 0;
	MPI_Sendrecv(rdata + k, offp_counts[j], type, dst, 111,
		     rdata + (nmax - k), offp_counts[l], type, src,
		     111, map->comm, &status);
	j = l;
	k = nmax - k;
	if (i == map->nprocs - 1)
	    break;
	/* pick up data from local data */
	for (l = 0; l < offp_counts[j]; l++) {
	    if ((m = rdata[k + l].gindex) < first)
		continue;
	    if (m >= last)
		continue;
	    rdata[k + l].value = data[m - first];
	}
    }
    MPI_Type_free(&type);
#endif /* USE_MPI */

  skip_comm:
    /* copy data to DOFs */
    for (l = 0; l < ndof; l++) {
	u = dofs[l];
	m = DofGetDataCount(u);
	for (i = 0; i < m; i++) {
	    j = map->L2Vmap[phgMapD2L(map, l, i)];
	    if (j < nlocal) {
		u->data[i] = data[j];
	    }
	    else {
		u->data[i] = rdata[k + j - nlocal].value;
	    }
	}
    }

    phgFree(offp_counts);
    phgFree(rdata);

    return 0;
  }
#endif	/*==================================================================*/
	
#if 0
  {
    int id;
    VEC *vec = phgMapCreateVec(map, 0);
    DOF ***pdofs = phgAlloc(ndof * sizeof(*pdofs));

    for (id = 0; id < ndof; id++)
	pdofs[id] = &dofs[id];
    vec->nvec = 1;
    vec->data = data;
    phgMapVecToDofArraysN(map, vec, FALSE, ndof, pdofs);
    vec->nvec = 0;
    vec->data = NULL;
    phgVecDestroy(&vec);
    phgFree(pdofs);
  }
#else	/* 0|1 */
  {
    int id;
    INT i, k, m;
    FLOAT *offp_data;
    DOF *u;

    assert(map->dofs != NULL);
    assert(ndof == map->ndof);

    offp_data = phgAlloc((map->localsize - map->nlocal) * sizeof(*data));

#if USE_MPI
    /* fetch remote data */
    if (map->nprocs > 1) {
	if (map->cinfo == NULL)
	    map->cinfo = phgMapCreateCommInfo(
				map->comm,
				map->nprocs,
				map->rank,
				map->nlocal,
				map->localsize,
				map->O2Gmap,
				map->ordering,
				map->partition);
	phgMapScatter(map->cinfo, 1, data, offp_data);
    }
#endif	/* USE_MPI */

    assert(map->localsize == 0 || map->L2Vmap != NULL);
    for (id = 0; id < ndof; id++, dofs++) {
	u = dofs[id];
	m = DofGetDataCount(u);
	for (i = 0; i < m; i++) {
	    k = map->L2Vmap[phgMapD2L(map, id, i)];
	    if (k < map->nlocal)
		u->data[i] = data[k];
	    else
		u->data[i] = offp_data[k - map->nlocal];
	}
    }

    phgFree(offp_data);
  }
#endif	/* 0|1 */

    return 0;
}

int
phgMapDofToLocalData(MAP *map, int ndof, DOF **dofs, FLOAT *data)
/* This function copies the data in the DOFs to local vector 
 * array (for e.g., setting up an initial solution of a solver).
 * 'data' is the pointer to the local data (size = map->nlocal)
 */
{
    DOF *u;
    int k;
    INT i, j, m;
    INT nlocal = map->nlocal;

    assert(map->localsize == 0 || map->L2Vmap != NULL);
    for (k = 0; k < ndof; k++) {
	u = dofs[k];
	m = DofGetDataCount(u);
	for (i = 0; i < m; i++) {
	    j = map->L2Vmap[phgMapD2L(map, k, i)];
	    if (j < nlocal) {
		data[j] = u->data[i];
	    }
	}
    }

    return 0;
}

VEC *
phgMapDofArraysToVecN(MAP *map, int nvec, BOOLEAN remove_bdry,
			 VEC **vecptr, int ndof, DOF ***dofs)
/* copies data from an array of DOFs to a vector */
{
    VEC *vec = (vecptr == NULL ? NULL : *vecptr);
    int id, iv;
    DOF *u, *u0;
    INT i, j, k, m, *table;
    FLOAT *data;

    assert(map->dofs != NULL);
    assert(ndof == map->ndof);
    MagicCheck(VEC, vec);

    if (vec != NULL) {
	phgVecDestroy(&vec);
	if (vecptr != NULL)
	    *vecptr = NULL;
    }

    if (nvec <= 0)
	return vec;

    /* expand bdry[] for faster checking */
    if (!remove_bdry) {
	/* no boundary entries */
	vec = phgMapCreateVec(map, nvec);
	table = phgAlloc(map->localsize * sizeof(*table));
	for (i = 0; i < map->localsize; i++)
	    table[i] = i;
    }
    else {
	MAP *map0 = phgMapRemoveBoundaryEntries(map);
	vec = phgMapCreateVec(map0, nvec);
	phgMapDestroy(&map0);
	table = phgCalloc(map->localsize, sizeof(*table));
	for (i = 0; i < map->bdry_localsize; i++)
	    table[map->bdry[i]] = 1;
	for (i = 0, k = 0; i < map->localsize; i++) {
	    if (table[i] == 1)
		table[i] = -1;
	    else
		table[i] = k++;
	}
    }

    if (vecptr != NULL)
	*vecptr = vec;

    assert(map->localsize == 0 || map->L2Vmap != NULL);
    for (id = 0; id < ndof; id++, dofs++) {
	u0 = (*dofs)[0];
	m = DofGetDataCount(u0);
	for (i = 0; i < m; i++) {
	    k = table[j = map->L2Vmap[phgMapD2L(map, id, i)]];
	    if (k == -1 || k >= vec->map->nlocal)
		continue;
	    data = vec->data + k;
	    for (iv = 0; iv < nvec; iv++) {
		u = (*dofs)[iv];
		assert(u->type == u0->type && u->dim == u0->dim);
		*data = u->data[i];
		data += vec->map->nlocal;
	    }
	}
    }
    phgFree(table);

    return vec;
}

VEC *
phgMapDofArraysToVec(MAP *map, int nvec, BOOLEAN remove_bdry,
			VEC **vecptr, DOF **u, ...)
/* copies data from an array of DOFs to a vector */
{
    int ndof;
    VEC *vec;
    DOF ***dofs;

    GetVariableArgs(ndof, dofs, u, DOF **);
    vec = phgMapDofArraysToVecN(map, nvec, remove_bdry, vecptr, ndof, dofs);
    phgFree(dofs);
    return vec;
}

void
phgMapVecToDofArraysN(MAP *map, VEC *vec, BOOLEAN remove_bdry,
			 int ndof, DOF ***dofs)
/* copies data from a vector to an array of DOFs */
{
    int id, iv, nvec;
    INT i, j, k, m, stride;
    FLOAT *data, *offp_data = NULL;
    static FLOAT zero = 0.0;
    DOF *u0, *u;
    INT *table;

    assert(map->dofs != NULL);
    assert(ndof == map->ndof);
    assert(vec != NULL);
    MagicCheck(VEC, vec);

    if (vec->nvec <= 0)
	return;

    if (!vec->assembled)
	phgVecAssemble(vec);

    nvec = vec->nvec;

    /* expand bdry[] for faster checking */
    if (!remove_bdry || vec->map->nglobal == map->nglobal) {
	/* no boundary entries */
	assert(vec->map->nglobal   == map->nglobal &&
	       vec->map->localsize == map->localsize);
	table = phgAlloc(map->localsize * sizeof(*table));
	for (i = 0; i < map->localsize; i++)
	    table[i] = i;
    }
    else {
	assert(vec->map->nlocal    == map->nlocal    - map->bdry_nlocal &&
	       vec->map->nglobal   == map->nglobal   - map->bdry_nglobal);
	table = phgCalloc(map->localsize, sizeof(*table));
	for (i = 0; i < map->bdry_localsize; i++)
	    table[map->bdry[i]] = 1;
	for (i = 0, k = 0; i < map->localsize; i++) {
	    if (table[i] == 1)
		table[i] = -1;
	    else
		table[i] = k++;
	}
    }

#if USE_MPI
    /* fetch remote data */
    if (map->nprocs > 1 && map->cinfo == NULL)
	map->cinfo = phgMapCreateCommInfo(
				map->comm,
				map->nprocs,
				map->rank,
				map->nlocal,
				map->localsize,
				map->O2Gmap,
				map->ordering,
				map->partition);

    if (map->cinfo != NULL && map->cinfo->comm != MPI_COMM_NULL) {
	/* collect off-process data */
	int ii;
	FLOAT *sbuff, *rbuff, *p, *q;
	MPI_Datatype type;
	COMM_INFO *cinfo = map->cinfo;

	if (nvec > 1) {
	    MPI_Type_contiguous(nvec, PHG_MPI_FLOAT, &type);
	    MPI_Type_commit(&type);
	}
	else {
	    type = PHG_MPI_FLOAT;
	}

	sbuff = phgAlloc(nvec * (cinfo->ssize + cinfo->rsize) * sizeof(*sbuff));
	rbuff = sbuff + nvec * cinfo->ssize;

	/* copy data to send from data to sbuff */
	p = sbuff;
	for (i = 0; i < cinfo->ssize; i++) {
	    k = table[cinfo->sind[i]];
	    assert(k < vec->map->nlocal);
	    if (k == -1) {
		q = &zero;
		stride = 0;
	    }
	    else {
		q = vec->data + k;
		stride = vec->map->nlocal;
	    }
	    for (ii = 0; ii < nvec; ii++) {
		*(p++) = *q;
		q += stride;
	    }
	}

	phgAlltoallv(sbuff, cinfo->scnts, cinfo->sdsps, type,
		     rbuff, cinfo->rcnts, cinfo->rdsps, type, cinfo->comm);

	q = rbuff;
	offp_data = phgAlloc(nvec * cinfo->rsize * sizeof(*offp_data));
	for (i = 0; i < cinfo->rsize; i++) {
	    p = offp_data + cinfo->rind[i];
	    for (ii = 0; ii < nvec; ii++) {
		*p = *(q++);
		p += cinfo->rsize;
	    }
	}

	phgFree(sbuff);
	if (nvec > 1)
	    MPI_Type_free(&type);
    }
#endif	/* USE_MPI */

    assert(map->localsize == 0 || map->L2Vmap != NULL);
    for (id = 0; id < ndof; id++, dofs++) {
	u0 = (*dofs)[0];
	m = DofGetDataCount(u0);
	for (i = 0; i < m; i++) {
	    k = table[j = map->L2Vmap[phgMapD2L(map, id, i)]];
	    if (k == -1) {
		data = &zero;
		stride = 0;
	    }
	    else if (k < vec->map->nlocal) {
		data = vec->data + k;
		stride = vec->map->nlocal;
	    }
	    else {
		assert(j >= map->nlocal);
		data = offp_data + j - map->nlocal;
		stride = map->localsize - map->nlocal;
	    }
	    for (iv = 0; iv < nvec; iv++) {
		u = (*dofs)[iv];
		assert(u->type == u0->type && u->dim == u0->dim);
		u->data[i] = *data;
		data += stride;
	    }
	}
    }

    phgFree(offp_data);
    phgFree(table);
}

void
phgMapVecToDofArrays(MAP *map, VEC *vec, BOOLEAN remove_bdry, DOF **u, ...)
/* copies data from a vector to an array of DOFs */
{
    int ndof;
    DOF ***dofs;

    GetVariableArgs(ndof, dofs, u, DOF **);
    phgMapVecToDofArraysN(map, vec, remove_bdry, ndof, dofs);
    phgFree(dofs);
}

VEC *
phgMapCreateVec(MAP *map, int nvec)
{
    VEC *v = phgCalloc(1, sizeof(*v));

    map->ref_count++;

    v->magic = MAGIC_VEC;
    v->map = map;
    v->mat = NULL;

#if USE_MPI
    v->O2Gmap = NULL;
    v->ordering = NULL;
    v->localsize = 0;
    v->alloc = 0;
#endif	/* USE_MPI */

    v->assembled = TRUE;
    v->nvec = nvec;
    v->data = phgCalloc(nvec * map->nlocal, sizeof(*(v->data)));
    v->offp_data = phgCalloc(nvec * (map->localsize - map->nlocal),
						sizeof(*(v->data)));

    return v;
}

MAT *
phgMapCreateMat(MAP *rmap, MAP *cmap)
{
    MAT *Mp;

    FunctionEntry;

    Mp = phgCalloc(1, sizeof(*Mp));
    Mp->magic = MAGIC_MAT;
    Mp->rmap = rmap;
    Mp->cmap = cmap;
    Mp->type = PHG_UNPACKED;

    /* check for special matrix */
    if (Mp->rmap->nprocs == 1 && Mp->cmap->nprocs > 1) {
	phgPrintf("\n*\n* Warning: Matrices with serial rows and distributed "
		  "columns are inefficient.\n");
	phgPrintf("* Warning: It is better to use distributed rows and serial "
		  "columns instead.\n*\n");
    }

#if USE_MPI
    Mp->O2Gmap = NULL;
    Mp->ordering = NULL;
    Mp->cinfo = NULL;
#endif	/* USE_MPI */
    Mp->localsize = -1;	/* to make convert_row_index() initialize Mp->O2Gmap,
			 * etc., when needed. */

    Mp->rows = phgCalloc(rmap->localsize, sizeof(*Mp->rows));
    Mp->nnz_d = Mp->nnz_o = 0;

    Mp->diag = NULL;
    Mp->packed_data = NULL;
    Mp->packed_cols = NULL;

    Mp->mv_func = NULL;
    Mp->mv_data = NULL;

    Mp->assembled = FALSE;
    Mp->mode = PHG_UNDEFINED;

    rmap->ref_count++;
    cmap->ref_count++;

    Return Mp;
}

MAT *
phgMapCreateMatrixFreeMat(MAP *rmap, MAP *cmap, MV_FUNC mv_func,
			  void *mv_data, ...)
{
    extern void _phg_matrix_free_setup(MAT *, MV_FUNC, int, void **);
    MAT *mat = phgMapCreateMat(rmap, cmap);
    int ndata;
    void **data;

    if (mat == NULL)
	return NULL;

    GetVariableArgs(ndata, data, mv_data, void *);
    _phg_matrix_free_setup(mat, mv_func, ndata, data);
    phgFree(data);

    return mat;
}
