/* Parallel Hierarchical Grid -- an adaptive finite element library.
 *
 * Copyright (C) 2005-2008 State Key Laboratory of Scientific and
 * Engineering Computing, Chinese Academy of Sciences. */

/* This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA */

/* Medit (http://www-rocq.inria.fr/gamma/medit/medit.html)
 * "mesh" file format import/export.
 *
 * $Id: medit.c,v 1.49 2008/01/30 08:57:37 zlb Exp $
*/

#include "phg.h"
#include "tetrahedralize.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

static BOOLEAN
get_token(FILE *fp, char *token)
{
    int c;
    char *p;

    while (TRUE) {
	if (fscanf(fp, "%s", token) != 1) return FALSE;
	if (token[0] != '#') break;
	/* skip to newline */
	do
	    if ((c = fgetc(fp)) == EOF) return FALSE;
	while (c != '\n');
    }
    if ((p = strchr(token, '#')) != NULL) *p = '\0';
    return TRUE;
}

static int
default_bcmap(int bc)
{
    switch (bc) {
	case 1:         return DIRICHLET;
	case 2:         return NEUMANN;
	case 3:         return BDRY_USER1;
	case 4:         return BDRY_USER2;
	default:        return -1;              /* invalid bctype */
    }
}

/* list of input_bc's which are mapped to UNDEFINED */
static INT *bc_list = NULL, bc_alloc = 0, bc_n = 0;

static int
bc_map(int input_bc)
{
    int i, j, bc = UNDEFINED;
    INT k = input_bc;
    INT *found;

    if (bc_n == 0)
	found = NULL;
    else
	found = bsearch(&k, bc_list, bc_n, sizeof(*bc_list), phgCompINT);

    if (phgBdryMapFunc_ == NULL)
	bc = default_bcmap(input_bc);
    else
	bc = (*phgBdryMapFunc_)(input_bc);

    if (found == NULL) {
	if (bc_n >= bc_alloc)
	    bc_list = phgRealloc(bc_list, (bc_alloc += 16) * sizeof(*bc_list));
	for (i = 0; i < bc_n; i++)
	    if (bc_list[i] > input_bc)
		break;
	for (j = bc_n; j > i; j--)
	    bc_list[j] = bc_list[j - 1];
	bc_list[i] = input_bc;
	bc_n++;
	if (bc < 0)
	    phgWarning("bdry type %d mapped to UNDEFINED\n", input_bc);
	else
	    phgInfo(1, "bdry type %d mapped to %s\n", input_bc, BTypeName(bc));
    }

    return bc < 0 ? UNDEFINED : bc;
}

static int
comp_face3(const void *p0, const void *p1)
/* compares two triangular faces */
{
    INT (*f0)[4] = (void *)p0, (*f1)[4] = (void *)p1;
    int i;

    if ((i = (*f0)[0] - (*f1)[0]) != 0)
	return i;
    if ((i = (*f0)[1] - (*f1)[1]) != 0)
	return i;
    return (*f0)[2] - (*f1)[2];
}

static void
process_tetrahedra(GRID *g, INT nlist, INT (*list)[4], INT *vtypes)
/* assign boundary types (e->bound_type[]) using the list of triangles */
{
    /* Note: each face is stored as (v0, v1, v2, elem_no, face_no) */
    INT i;
    INT key[4], (*found)[4];
    SIMPLEX *e;
    int j, elem_type;

    if (nlist > 0)
	qsort(list, nlist, sizeof(*list), comp_face3);

    for (i = 0; i < g->nroot; i++) {
	e = g->roots + i;
	/* element type was saved in mark */
	elem_type = (e->mark == 0 ? UNDEFINED : bc_map(e->mark));
	for (j = 0; j < NFace; j++) {
#if 0

/* TODO: should first check whether the face is a boundary face before
	 setting its type to the type of the element or vertices,
	 because if an interior face is set a type other than UNDEFINED,
	 it will be considered as a boundary face later. */

	    /* first, set bdry type to element type */
	    e->bound_type[j] = elem_type;
	    /* then, adjust bdry type according to types of vertices */
	    key[0] = e->verts[GetFaceVertex(j, 0)];
	    key[1] = e->verts[GetFaceVertex(j, 1)];
	    key[2] = e->verts[GetFaceVertex(j, 2)];
	    if (vtypes != NULL) {
		int k;
		if ((k = vtypes[key[0]]) != 0 &&
		    k == vtypes[key[1]] && k == vtypes[key[2]])
		    e->bound_type[j] = bc_map(k);
	    }
#else
	    (void)elem_type;
	    key[0] = e->verts[GetFaceVertex(j, 0)];
	    key[1] = e->verts[GetFaceVertex(j, 1)];
	    key[2] = e->verts[GetFaceVertex(j, 2)];
	    e->bound_type[j] = UNDEFINED;
#endif
	    /* finally, adjust bdry type according to types of triangles */
	    if (nlist == 0)
		continue;
	    qsort(key, 3, sizeof(INT), phgCompINT);
	    found = bsearch(key, list, nlist, sizeof(*list), comp_face3);
	    if (found != NULL && (*found)[3] != 0)
		e->bound_type[j] = bc_map((*found)[3]);
	}
    }
}

static int
comp_face4(const void *p0, const void *p1)
/* compares two triangular faces */
{
    INT (*f0)[5] = (void *)p0, (*f1)[5] = (void *)p1;
    int i;

    if ((i = (*f0)[0] - (*f1)[0]) != 0)
	return i;
    if ((i = (*f0)[1] - (*f1)[1]) != 0)
	return i;
    if ((i = (*f0)[2] - (*f1)[2]) != 0)
	return i;
    return (*f0)[3] - (*f1)[3];
}

static INT ntetra, ntetra_allocated;
static SIMPLEX *tetras;

static void
new_tetra(int v0, int v1, int v2, int v3, int bound_type[])
{
    SIMPLEX *e;
    int i, b;

    if (v0 == v1 || v0 == v2 || v0 == v3 || v1 == v2 || v1 == v3 || v2 == v3) {
	phgWarning("deleting invalid tetrahedron: %d %d %d %d\n", v0,v1,v2,v3);
	return;
    }
    if (ntetra >= ntetra_allocated) {
	ntetra_allocated += 1024;
	tetras = phgReallocElements(tetras, ntetra, ntetra_allocated);
    }

    e = tetras + (ntetra++);
    e->verts[0] = v0;
    e->verts[1] = v1;
    e->verts[2] = v2;
    e->verts[3] = v3;
    if (bound_type != NULL) {
	for (i = 0; i < NVert; i++) {
	    b = bound_type[i];
	    if (b >= 0 && b != UNDEFINED)
		e->bound_type[i] = b;
	}
    }
}

static void
process_hexahedra(GRID *g, int nhexa, int (*hexas)[9],
		  INT nlist, INT (*list)[5], INT *vtypes)
/* assign boundary types (e->bound_type[]) using the list of triangles */
{
    /* Note: each face is stored as (v0, v1, v2, elem_no, face_no) */
    INT i;
    INT key[5], (*found)[5];
    int j, bound_type[6], elem_type;

    if (nlist > 0)
	qsort(list, nlist, sizeof(*list), comp_face4);

    ntetra = ntetra_allocated = g->nroot;
    tetras = g->roots;
    for (i = 0; i < nhexa; i++) {
	elem_type = (hexas[i][8] == 0 ? UNDEFINED : bc_map(hexas[i][8]));
	for (j = 0; j < 6; j++) {
#if 0

/* TODO: should first check whether the face is a boundary face before
	 setting its type to the type of the element or vertices,
	 because if an interior face is set a type other than UNDEFINED,
	 it will be considered as a boundary face later. */

	    /* first, set bdry type to element type */
	    bound_type[j] = elem_type;
	    /* then, adjust bdry type according to types of vertices */
	    key[0] = hexas[i][phgElemInfo[BRICK].face_info[j][1]];
	    key[1] = hexas[i][phgElemInfo[BRICK].face_info[j][2]];
	    key[2] = hexas[i][phgElemInfo[BRICK].face_info[j][3]];
	    key[3] = hexas[i][phgElemInfo[BRICK].face_info[j][4]];
	    if (vtypes != NULL) {
		int k;
		if ((k = vtypes[key[0]]) != 0 &&
		    k == vtypes[key[1]] && k == vtypes[key[2]] &&
		    k == vtypes[key[3]])
		    bound_type[j] = bc_map(k);
	    }
#else
	    Unused(elem_type);
	    key[0] = hexas[i][phgElemInfo[BRICK].face_info[j][1]];
	    key[1] = hexas[i][phgElemInfo[BRICK].face_info[j][2]];
	    key[2] = hexas[i][phgElemInfo[BRICK].face_info[j][3]];
	    key[3] = hexas[i][phgElemInfo[BRICK].face_info[j][4]];
	    bound_type[j] = UNDEFINED;
#endif
	    /* finally, adjust bdry type according to types of quadrilaterals */
	    if (nlist == 0)
		continue;
	    qsort(key, 4, sizeof(INT), phgCompINT);
	    found = bsearch(key, list, nlist, sizeof(*list), comp_face4);
	    if (found != NULL && (*found)[4] != 0)
		bound_type[j] = bc_map((*found)[4]);
	}
	/* convert to tetrahedra */
	phgBrick2Tetra(hexas[i], new_tetra, bound_type);
    }
    phgInfo(2, "%d hexahedr%s converted to %d tetrahedra\n",
		nhexa, nhexa > 1 ? "a" : "on", ntetra - g->nroot);
    g->ntree = g->nleaf = g->nelem = g->nroot = g->nelem_global = ntetra;
    g->roots = phgRealloc(tetras, ntetra * sizeof(*g->roots));
}

int
phgImportMedit(GRID *g, const char *filename, BOOLEAN parallel)
{
    FILE *fp;
    char token[1024]; 
    int i, n;
    INT *vtypes = NULL;			/* vertex types */
    INT nlist3 = 0, (*list3)[4] = NULL;	/* list of triangles */
    INT nlist4 = 0, (*list4)[5] = NULL;	/* list of quadrilaterals */
    INT nlistt = 0;			/* number of terahedra */
    int nlisth = 0, (*listh)[9] = NULL;	/* list of hexahedra */

#define READ_NUMBER						\
    if (!get_token(fp, token)) strcpy(token, "End");		\
    if (isalpha((int)(token[0]))) {				\
	phgWarning("fewer entries (%d) than expected.\n", i);	\
	break;							\
    }

    FunctionEntry;

    (void)(parallel);	/* avoid gcc warning */

    if (phgRank > 0)
	Return 0;

    if ((fp = phgOpenInputFile_(filename)) == NULL) {
	phgWarning("can't open mesh file \"%s\"!\n", filename);
	phgFreeGrid(&g);
	Return -1;
    }

    phgInfo(1, "loading file \"%s\".\n", filename);

    if (!get_token(fp, token) || strcasecmp(token, "MeshVersionFormatted") ||
	!get_token(fp, token) || strcmp(token, "1") ||
	!get_token(fp, token) || strcasecmp(token, "Dimension") ||
	!get_token(fp, token) || strcmp(token, "3")) {
error:
	phgError(1, "invalid or unsupported input file, abort.\n");
    }

    while (TRUE) {
	if (!get_token(fp, token))
	    break;
next_token:
	if (!strcasecmp(token, "End")) {
	    break;
	}
	else if (!strcasecmp(token, "Vertices")) {
	    BOOLEAN flag = FALSE;
	    if (!get_token(fp, token))
		goto error;
	    n = atoi(token);
	    phgInfo(2, "number of vertices: %d\n", n);
	    g->verts = phgNewVertices(n);
	    vtypes = phgAlloc(n * sizeof(*vtypes)); 
	    for (i = 0; i < n; i++) {
		READ_NUMBER
		g->verts[i][0] = atof(token);
		if (!get_token(fp, token))
		    goto error;
		g->verts[i][1] = atof(token);
		if (!get_token(fp, token))
		    goto error;
		g->verts[i][2] = atof(token);
		/* the type of the vertex */
		if (!get_token(fp, token))
		    goto error;
		if ((vtypes[i] = atoi(token)) != 0)
		    flag = TRUE;
	    }
	    g->nvert_global = g->nvert = i;
	    if (!flag) {
		phgFree(vtypes);
		vtypes = NULL;
	    }
	    if (i < n)
		goto next_token;
	}
	else if (!strcasecmp(token, "Edges")) {
	    if (!get_token(fp, token))
		goto error;
	    n = atoi(token);
	    phgInfo(2, "number of edges: %d (ignored)\n", n);
	    for (i = 0; i < n; i++) {
		READ_NUMBER
		if (!get_token(fp, token))
		    goto error;
		if (!get_token(fp, token))
		    goto error;
	    }
	    if (i < n)
		goto next_token;
	}
	else if (!strcasecmp(token, "Triangles")) {
	    if (!get_token(fp, token))
		goto error;
	    /* extract boundary types from list of triangles */
	    n = atoi(token);
	    phgInfo(2, "number of triangles: %d\n", n);
	    list3 = phgRealloc(list3, (n + nlist3) * sizeof(*list3));
	    for (i = 0; i < n; i++) {
		READ_NUMBER
		list3[nlist3][0] = atoi(token) - 1;
		if (!get_token(fp, token))
		    goto error;
		list3[nlist3][1] = atoi(token) - 1;
		if (!get_token(fp, token))
		    goto error;
		list3[nlist3][2] = atoi(token) - 1;
		if (!get_token(fp, token))
		    goto error;
		list3[nlist3][3] = atoi(token);	/* boundary type */
		/* ignore boundary type 0 */
		if (list3[nlist3][3] == 0)
		    continue;
		/* sort the three vertices */
		qsort(list3[nlist3++], 3, sizeof(INT), phgCompINT);
	    }
	    if (nlist3 == 0) {
		phgFree(list3);
		list3 = NULL;
	    }
	    if (i < n)
		goto next_token;
	}
	else if (!strcasecmp(token, "Quadrilaterals")) {
	    if (!get_token(fp, token))
		goto error;
	    n = atoi(token);
	    phgInfo(2, "number of quadrilaterals: %d\n", n);
	    list4 = phgRealloc(list4, (n + nlist4) * sizeof(*list4));
	    for (i = 0; i < n; i++) {
		READ_NUMBER
		list4[nlist4][0] = atoi(token) - 1;
		if (!get_token(fp, token))
		    goto error;
		list4[nlist4][1] = atoi(token) - 1;
		if (!get_token(fp, token))
		    goto error;
		list4[nlist4][2] = atoi(token) - 1;
		if (!get_token(fp, token))
		    goto error;
		list4[nlist4][3] = atoi(token) - 1;
		if (!get_token(fp, token))
		    goto error;
		list4[nlist4][4] = atoi(token);	/* boundary type */
		/* ignore boundary type 0 */
		if (list4[nlist4][4] == 0)
		    continue;
		/* sort the four vertices */
		qsort(list4[nlist4++], 4, sizeof(INT), phgCompINT);
	    }
	    if (nlist4 == 0) {
		phgFree(list4);
		list4 = NULL;
	    }
	    if (i < n)
		goto next_token;
	}
	else if (!strcasecmp(token, "Tetrahedra")) {
	    if (!get_token(fp, token))
		    goto error;
	    n = atoi(token);
	    phgInfo(2, "number of tetrahedra: %d\n", n);
	    g->roots = phgReallocElements(g->roots, nlistt, n + nlistt);
	    for (i = 0; i < n; i++) {
		READ_NUMBER
		g->roots[nlistt].verts[0] = atoi(token) - 1;
		if (!get_token(fp, token))
		    goto error;
		g->roots[nlistt].verts[1] = atoi(token) - 1;
		if (!get_token(fp, token))
		    goto error;
		g->roots[nlistt].verts[2] = atoi(token) - 1;
		if (!get_token(fp, token))
		    goto error;
		g->roots[nlistt].verts[3] = atoi(token) - 1;
		/* save type of tetrahedron in 'mark' (currently not used) */
		if (!get_token(fp, token))
		    goto error;
		g->roots[nlistt++].mark = atoi(token);
	    }
	    if (i < n)
		goto next_token;
	}
	else if (!strcasecmp(token, "Hexahedra")) {
	    if (!get_token(fp, token))
		goto error;
	    n = atoi(token);
	    phgInfo(2, "number of hexahedra: %d\n", n);
	    listh = phgRealloc(listh, (n + nlisth) * sizeof(*listh));
	    for (i = 0; i < n; i++) {
		READ_NUMBER
		listh[nlisth][0] = atoi(token) - 1;
		if (!get_token(fp, token))
		    goto error;
		listh[nlisth][1] = atoi(token) - 1;
		if (!get_token(fp, token))
		    goto error;
		listh[nlisth][2] = atoi(token) - 1;
		if (!get_token(fp, token))
		    goto error;
		listh[nlisth][3] = atoi(token) - 1;
		if (!get_token(fp, token))
		    goto error;
		listh[nlisth][4] = atoi(token) - 1;
		if (!get_token(fp, token))
		    goto error;
		listh[nlisth][5] = atoi(token) - 1;
		if (!get_token(fp, token))
		    goto error;
		listh[nlisth][6] = atoi(token) - 1;
		if (!get_token(fp, token))
		    goto error;
		listh[nlisth][7] = atoi(token) - 1;
		/* The element type */
		if (!get_token(fp, token))
		    goto error;
		listh[nlisth++][8] = atoi(token);
	    }
	    if (i < n)
		goto next_token;
	}
	else if (!strcasecmp(token, "Corners")) {
	    if (!get_token(fp, token))
		    goto error;
	    n = atoi(token);
	    for (i = 0; i < n; i++) {
		READ_NUMBER
	    }
	    if (i < n)
		goto next_token;
	}
	else if (!strcasecmp(token, "RequiredVertices")) {
	    if (!get_token(fp, token))
		    goto error;
	    n = atoi(token);
	    for (i = 0; i < n; i++) {
		READ_NUMBER
	    }
	    if (i < n)
		goto next_token;
	}
	else if (!strcasecmp(token, "Ridges")) {
	    if (!get_token(fp, token))
		    goto error;
	    n = atoi(token);
	    for (i = 0; i < n; i++) {
		READ_NUMBER
	    }
	    if (i < n)
		goto next_token;
	}
	else if (!strcasecmp(token, "RequiredEdges")) {
	    if (!get_token(fp, token))
		    goto error;
	    n = atoi(token);
	    for (i = 0; i < n; i++) {
		READ_NUMBER
	    }
	    if (i < n)
		goto next_token;
	}
	else {
#if 0
	    int lineno = 1;
	    size_t pos, current_pos = ftell(fp);
	    fseek(fp, 0, SEEK_SET);
	    for (pos = 0; pos < current_pos; pos++)
		if (fgetc(fp) == '\n')
		    lineno++;
	    phgCloseInputFile_(fp);
	    phgError(1, "unkown token \"%s\" near line %d.\n", token, lineno);
#else
	    /* tetgen 1.4.0 seems to generate inconsistent (smaller) number of
	     * triangles, as a workaround extra numbers are silently ignored */
	    static BOOLEAN warned = FALSE;
	    if (!warned) {
		int lineno = 1;
		size_t pos, current_pos = ftell(fp);
		warned = TRUE;
		fseek(fp, 0, SEEK_SET);
		for (pos = 0; pos < current_pos; pos++)
		    if (fgetc(fp) == '\n')
			lineno++;
		phgWarning("extra token \"%s\" near line %d ignored.\n",
				token, lineno);
	    }
	    continue;
#endif
	}
    }

    phgCloseInputFile_(fp);

    g->nelem_global = g->ntree = g->nleaf = g->nelem = g->nroot = nlistt;

    if (g->nroot > 0)
	process_tetrahedra(g, nlist3, list3, vtypes);

    if (list3 != NULL) {
	phgFree(list3);
	list3 = NULL;
	nlist3 = 0;
    }

    if (listh != NULL) {
	process_hexahedra(g, nlisth, listh, nlist4, list4, vtypes);
	phgFree(listh);
	listh = NULL;
	nlisth = 0;
    }

    if (list4 != NULL) {
	phgFree(list4);
	list4 = NULL;
	nlist4 = 0;
    }

    if (vtypes != NULL) {
	phgFree(vtypes);
	vtypes = NULL;
    }

    phgFree(bc_list);
    bc_list = NULL;
    bc_alloc = bc_n = 0;

    Return 0;
    return 0;	/* for avoiding MIPS cc warning */
}

static FILE *f;

static BOOLEAN
elements_callback CB_ARGS(s)
{
#if Dim == 2
    fprintf(f, " %d %d %d 0\n", s->verts[0] + 1, s->verts[1] + 1,
		s->verts[2] + 1);
#else
    fprintf(f, " %d %d %d %d 0\n", s->verts[0] + 1,
		s->verts[1] + 1, s->verts[2] + 1, s->verts[3] + 1);
#endif
    return TRUE;
}

BOOLEAN
phgExportMedit(GRID *g, const char *filename)
/* Output Medit mesh file. filename == NULL => output to stdout. */
{
    INT i;

    FunctionEntry;

    if (g == NULL || g->nleaf == 0) Return TRUE;

    if (g->nprocs > 1) {
	if (g->rank == 0)
	    phgWarning("only the first submesh can be saved.\n");
	else
	    Return TRUE;
    }

    if (filename == NULL) {
	f = stdout;
    }
    else {
	if ((f = fopen(filename, "w+t")) == NULL) {
	    phgError(0, "cannot open output file \"%s\".\n", filename);
	    Return FALSE;
	}
	phgInfo(1, "creating Medit mesh file \"%s\".\n", filename);
    }

    fprintf(f, "MeshVersionFormatted 1\n");
    fprintf(f, "Dimension %d\n", Dim);
    fprintf(f, "\nVertices %d\n", g->nvert);
    for (i = 0; i < g->nvert; i++) {
#if Dim == 2
	fprintf(f, " %0.16lg %0.16lg 0\n",
			(double)g->verts[i][0], (double)g->verts[i][1]);
#else
	fprintf(f, " %0.16lg %0.16lg %0.16lg 0\n", (double)g->verts[i][0],
			(double)g->verts[i][1], (double)g->verts[i][2]);
#endif
    }

    fprintf(f, "\nTetrahedra %d\n", g->nleaf);
    phgTraverseElements(g, elements_callback);

    fprintf(f, "\nEnd\n");

    ((f != stdout && f != stderr) ? fclose : fflush)(f);

    Return TRUE;
    return TRUE;	/* for avoiding MIPS cc warning */
}
